# pyimpspec is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 pyimpspec developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of pyimpspec's dependencies and/or sources of portions of code are included in
# the LICENSES folder.

import cmath
from collections import OrderedDict
from math import pi
from os.path import (
    basename,
    splitext,
)
from typing import (
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)
from uuid import uuid4
from numpy import (
    allclose,
    angle,
    array,
    flip,
    integer,
    issubdtype,
    mean,
    ndarray,
    radians,
    bool_,
)
from pandas import DataFrame


VERSION: int = 1


def _parse_v1(dictionary: dict) -> dict:
    assert type(dictionary) is dict
    assert "frequency" in dictionary
    assert "real" in dictionary
    assert "imaginary" in dictionary
    return {
        "frequency": array(dictionary["frequency"]),
        "impedance": array(
            list(
                map(
                    lambda _: complex(*_),
                    zip(dictionary["real"], dictionary["imaginary"]),
                )
            )
        ),
        "mask": {int(k): v for k, v in dictionary.get("mask", {}).items()},
        "path": dictionary.get("path", ""),
        "label": dictionary.get("label", ""),
        "uuid": dictionary.get("uuid", ""),
    }


class DataSet:
    """
    A class that represents an impedance spectrum.
    The data points can be masked, which results in those data points being omitted from any analyses and visualization.

    Parameters
    ----------
    frequency: ndarray
        A 1-dimensional array of frequencies in hertz.

    impedance: ndarray
        A 1-dimensional array of complex impedances in ohms.

    mask: Dict[int, bool] = {}
        A mapping of integer indices to boolean values where a value of True means that the data point is to be omitted.

    path: str = ""
        The path to the file that has been parsed to generate this DataSet instance.

    label: str = ""
        The label assigned to this DataSet instance.

    uuid: str = ""
        The universally unique identifier assigned to this DataSet instance.
        If empty, then one will be automatically assigned.
    """

    def __init__(
        self,
        frequency: ndarray,
        impedance: ndarray,
        mask: Dict[int, bool] = {},
        path: str = "",
        label: str = "",
        uuid: str = "",
    ):
        assert type(frequency) is ndarray, type(frequency)
        assert type(impedance) is ndarray, type(impedance)
        assert frequency.shape == impedance.shape, (
            frequency.shape,
            impedance.shape,
        )
        assert (
            len(frequency) == len(impedance) > 0
        ), f"{path=} ({label=}): {len(frequency)=}, {len(impedance)=}"
        assert type(mask) is dict
        assert all(
            map(
                lambda _: issubdtype(type(_[0]), integer) and type(_[1]) is bool,
                mask.items(),
            )
        ), mask
        assert type(path) is str
        assert type(label) is str
        assert type(uuid) is str
        # Sort data points in descending order of frequency.
        if frequency[-1] > frequency[0]:
            frequency = flip(frequency)
            impedance = flip(impedance)
            if len(mask) > 0:
                i: int
                for i in range(0, frequency.size):
                    j: int = frequency.size - 1 - i
                    flag: bool = mask.get(i, False)
                    mask[i] = mask.get(j, False)
                    mask[j] = flag
        self.uuid: str = uuid or uuid4().hex
        self._path: str = path
        self._label: str = label or splitext(basename(path))[0]
        self._frequency: ndarray = frequency
        self._impedance: ndarray = impedance
        self._mask: Dict[int, bool] = mask or {
            i: False for i in range(0, len(frequency))
        }
        self._num_points: int = len(frequency)
        self.set_mask(mask)

    def __repr__(self) -> str:
        return f"DataSet ({self._label}, {hex(id(self))})"

    def subtract_impedance(self, Z: Union[complex, List[complex], ndarray]):
        """
        Subtract either the same complex value from all data points or a unique complex value for each data point in this DataSet.

        Parameters
        ----------
        Z: Union[complex, List[complex], ndarray]
            The complex value(s) to subtract from this DataSet's impedances.
        """
        if type(Z) is list or type(Z) is ndarray:
            assert all(map(lambda _: hasattr(_, "real") and hasattr(_, "imag"), Z)), (
                type(Z),
                type(Z[0]),
                Z[0],
            )
        else:
            assert hasattr(Z, "real") and hasattr(Z, "imag"), (
                type(Z),
                Z,
            )
        self._impedance = self._impedance - Z

    @staticmethod
    def _parse(dictionary: dict) -> dict:
        assert type(dictionary) is dict
        parsers: Dict[int, Callable] = {
            1: _parse_v1,
        }
        version: int = dictionary.get("version", VERSION)
        assert version <= VERSION, f"Unsupported version: {version=} > {VERSION=}"
        assert (
            version in parsers
        ), f"Unsupported version: {version=} not in {parsers.keys()=}"
        return parsers[version](dictionary)

    @classmethod
    def from_dict(Class, dictionary: dict) -> "DataSet":
        """
        Create a DataSet from a dictionary.

        Parameters
        ----------
        dictionary: dict
            A dictionary containing at least the frequencies, and the real and the imaginary parts of the impedances.

        Returns
        -------
        DataSet
        """
        return Class(**Class._parse(dictionary))

    @classmethod
    def copy(Class, data: "DataSet", label: Optional[str] = None) -> "DataSet":
        """
        Create a copy of an existing DataSet.

        Parameters
        ----------
        data: DataSet
            The existing DataSet to make a copy of.

        label: Optional[str] = None
            The label that the copy should have.

        Returns
        -------
        DataSet
        """
        assert type(data) is Class
        assert type(label) is str or label is None
        dictionary: dict = data.to_dict()
        if label is not None:
            dictionary["label"] = label
        del dictionary["uuid"]
        return Class.from_dict(dictionary)

    @classmethod
    def average(Class, data_sets: List["DataSet"], label: str = "Average") -> "DataSet":
        """
        Create a DataSet by averaging the impedances of multiple DataSet instances.

        Parameters
        ----------
        data_sets: List[DataSet]
            The DataSet instances to average.

        label: str = "Average"
            The label that the new DataSet should have.

        Returns
        -------
        DataSet
        """
        assert type(data_sets) is list and all(
            map(lambda _: type(_) is Class, data_sets)
        )
        assert type(label) is str
        freqs: List[ndarray] = list(
            map(lambda _: _.get_frequency(masked=None), data_sets)
        )
        imps: List[ndarray] = list(
            map(lambda _: _.get_impedance(masked=None), data_sets)
        )
        f: ndarray = freqs.pop(0)
        assert all(map(lambda _: allclose(f, _), freqs))
        Z: ndarray = mean(array(imps), axis=0)
        return Class.from_dict(
            {
                "frequency": f,
                "real": Z.real,
                "imaginary": Z.imag,
                "label": label,
            }
        )

    def get_path(self) -> str:
        """
        Get the path to the file that was parsed to generate this DataSet.

        Returns
        -------
        str
        """
        return self._path

    def set_path(self, path: str):
        """
        Set the path to the file that was parsed to generate this DataSet.

        Parameters
        ----------
        path: str
            The path.
        """
        assert type(path) is str
        self._path = path

    def get_label(self) -> str:
        """
        Get the label assigned to this DataSet.

        Returns
        -------
        str
        """
        return self._label

    def set_label(self, label: str):
        """
        Set the label assigned to this DataSet.

        Parameters
        ----------
        label: str
            The new label.
        """
        assert type(label) is str
        self._label = label

    def set_mask(self, mask: Dict[int, bool]):
        """
        Set the mask for this DataSet.

        Parameters
        ----------
        mask: Dict[int, bool]
            The new mask.
            The keys must be zero-based indices and the values must be boolean values.
            True means that the data point is to be omitted and False means that the data point is to be included.
        """
        assert (
            type(mask) is dict
            and all(map(lambda _: issubdtype(type(_), integer), mask.keys()))
            and all(map(lambda _: issubdtype(type(_), bool_), mask.values()))
        ), mask
        mask = mask.copy()
        i: int
        flag: bool
        for i, flag in mask.items():
            assert issubdtype(type(i), integer), type(i)
            assert issubdtype(type(flag), bool_), type(flag)
        for i in list(mask.keys()):
            if i < 0 or i >= self._num_points:
                del mask[i]
        self._mask = mask.copy()
        for i in range(0, self._num_points):
            if i not in self._mask:
                self._mask[i] = False

    def get_mask(self) -> Dict[int, bool]:
        """
        Get the mask for this DataSet.
        The keys are zero-based indices and the values are booleans.
        True means that the data point is to be omitted and False means that the data point is to be included.

        Returns
        -------
        Dict[int, bool]
        """
        return self._mask.copy()

    def get_frequency(self, masked: Optional[bool] = False) -> ndarray:
        """
        Get the frequencies in this DataSet.

        Parameters
        ----------
        masked: Optional[bool] = False
            None means that all frequencies are returned.
            True means that only frequencies that are to be omitted are returned.
            False means that only frequencies that are to be included are returned.

        Returns
        -------
        ndarray
        """
        assert type(masked) is bool or masked is None
        if masked is None:
            return array(self._frequency)
        i: int
        f: float
        return array(
            [
                f
                for i, f in enumerate(self._frequency)
                if self._mask.get(i, False) == masked
            ]
        )

    def get_impedance(self, masked: Optional[bool] = False) -> ndarray:
        """
        Get the complex impedances in this DataSet.

        Parameters
        ----------
        masked: Optional[bool] = False
            None means that all impedances are returned.
            True means that only impedances that are to be omitted are returned.
            False means that only impedances that are to be included are returned.

        Returns
        -------
        ndarray
        """
        assert type(masked) is bool or masked is None
        if masked is None:
            return self._impedance
        i: int
        c: complex
        return array(
            [
                c
                for i, c in enumerate(self._impedance)
                if self._mask.get(i, False) == masked
            ]
        )

    def get_real(self, masked: Optional[bool] = False) -> ndarray:
        """
        Get the real parts of the impedances in this DataSet.

        Parameters
        ----------
        masked: Optional[bool] = False
            None means that all impedances are returned.
            True means that only impedances that are to be omitted are returned.
            False means that only impedances that are to be included are returned.

        Returns
        -------
        ndarray
        """
        return self.get_impedance(masked=masked).real

    def get_imaginary(self, masked: Optional[bool] = False) -> ndarray:
        """
        Get the imaginary parts of the impedances in this DataSet.

        Parameters
        ----------
        masked: Optional[bool] = False
            None means that all impedances are returned.
            True means that only impedances that are to be omitted are returned.
            False means that only impedances that are to be included are returned.

        Returns
        -------
        ndarray
        """
        return self.get_impedance(masked=masked).imag

    def get_magnitude(self, masked: Optional[bool] = False) -> ndarray:
        """
        Get the absolute magnitudes of the impedances in this DataSet.

        Parameters
        ----------
        masked: Optional[bool] = False
            None means that all impedances are returned.
            True means that only impedances that are to be omitted are returned.
            False means that only impedances that are to be included are returned.

        Returns
        -------
        ndarray
        """
        return abs(self.get_impedance(masked=masked))

    def get_phase(self, masked: Optional[bool] = False) -> ndarray:
        """
        Get the phase angles/shifts of the impedances in this DataSet in degrees.

        Parameters
        ----------
        masked: Optional[bool] = False
            None means that all impedances are returned.
            True means that only impedances that are to be omitted are returned.
            False means that only impedances that are to be included are returned.

        Returns
        -------
        ndarray
        """
        return angle(self.get_impedance(masked=masked), deg=True)

    def get_num_points(self, masked: Optional[bool] = False) -> int:
        """
        Get the number of data points in this DataSet

        Parameters
        ----------
        masked: Optional[bool] = False
            None means that all impedances are returned.
            True means that only impedances that are to be omitted are returned.
            False means that only impedances that are to be included are returned.

        Returns
        -------
        int
        """
        return len(self.get_impedance(masked=masked))

    def get_nyquist_data(
        self, masked: Optional[bool] = False
    ) -> Tuple[ndarray, ndarray]:
        """
        Get the data necessary to plot this DataSet as a Nyquist plot: the real and the negative imaginary parts of the impedances.

        Parameters
        ----------
        masked: Optional[bool] = False
            None means that all impedances are returned.
            True means that only impedances that are to be omitted are returned.
            False means that only impedances that are to be included are returned.

        Returns
        -------
        Tuple[ndarray, ndarray]
        """
        Z: ndarray = self.get_impedance(masked=masked)
        return (
            Z.real,
            -Z.imag,
        )

    def get_bode_data(
        self, masked: Optional[bool] = False
    ) -> Tuple[ndarray, ndarray, ndarray]:
        """
        Get the data necessary to plot this DataSet as a Bode plot: the frequencies, the absolute magnitudes of the impedances, and the negative phase angles/shifts of the impedances in degrees.

        Parameters
        ----------
        masked: Optional[bool] = False
            None means that all impedances are returned.
            True means that only impedances that are to be omitted are returned.
            False means that only impedances that are to be included are returned.

        Returns
        -------
        Tuple[ndarray, ndarray, ndarray]
        """
        f: ndarray = self.get_frequency(masked=masked)
        Z: ndarray = self.get_impedance(masked=masked)
        return (
            f,
            abs(Z),
            -angle(Z, deg=True),
        )

    def to_dict(self) -> dict:
        """
        Get a dictionary that represents this DataSet, can be used to serialize the DataSet (e.g. as a JSON file), and then used to recreate this DataSet.

        Returns
        -------
        dict
        """
        return {
            "version": VERSION,
            "path": self._path,
            "label": self._label,
            "frequency": list(self._frequency),
            "real": list(self._impedance.real),
            "imaginary": list(self._impedance.imag),
            "mask": self.get_mask(),
            "uuid": self.uuid,
        }

    def to_dataframe(
        self,
        masked: Optional[bool] = False,
        frequency_label: str = "f (Hz)",
        real_label: Optional[str] = "Zre (ohm)",
        imaginary_label: Optional[str] = "Zim (ohm)",
        magnitude_label: Optional[str] = "|Z| (ohm)",
        phase_label: Optional[str] = "phase angle (deg.)",
        negative_imaginary: bool = False,
        negative_phase: bool = False,
    ) -> DataFrame:
        """
        Create a pandas.DataFrame instance from this DataSet.

        Parameters
        ----------
        masked: Optional[bool] = False
            None means that all impedances are returned.
            True means that only impedances that are to be omitted are returned.
            False means that only impedances that are to be included are returned.

        frequency_label: str = "f (Hz)"
            The label assigned to the frequency data.

        real_label: Optional[str] = "Zre (ohm)"
            The label assigned to the real part of the impedance data.

        imaginary_label: Optional[str] = "Zim (ohm)"
            The label assigned to the imaginary part of the impedance data.

        magnitude_label: Optional[str] = "|Z| (ohm)"
            The label assigned to the magnitude of the impedance data.

        phase_label: Optional[str] = "phase angle (deg.)"
            The label assigned to the phase of the imedance data.

        negative_imaginary: bool = False
            Whether or not the sign of the imaginary part of the impedance data should be inverted.

        negative_phase: bool = False
            Whether or not the sign of the phase of the impedance data should be inverted.

        Returns
        -------
        DataFrame
        """
        assert type(frequency_label) is str
        assert type(real_label) is str or real_label is None
        assert type(imaginary_label) is str or imaginary_label is None
        assert type(magnitude_label) is str or magnitude_label is None
        assert type(phase_label) is str or phase_label is None
        assert type(negative_imaginary) is bool
        assert type(negative_phase) is bool
        dictionary: Dict[str, ndarray] = {
            frequency_label: self._frequency,
        }
        Z: ndarray = self.get_impedance(masked=masked)
        if real_label is not None:
            assert imaginary_label is not None
            dictionary[real_label] = Z.real
            dictionary[imaginary_label] = Z.imag * (-1 if negative_imaginary else 1)
        if magnitude_label is not None:
            assert phase_label is not None
            dictionary[magnitude_label] = abs(Z)
            dictionary[phase_label] = angle(Z, deg=True) * (-1 if negative_phase else 1)
        return DataFrame(dictionary)


def dataframe_to_dataset(df: DataFrame, path: str, label: str = "") -> DataSet:
    """
    Convert a pandas.DataFrame into a DataSet.

    Parameters
    ----------
    df: DataFrame
        The DataFrame to be converted.

    path: str
        The path to the file that was used to create the DataFrame.

    label: str = ""
        The label assigned to the new DataSet.

    Returns
    -------
    DataSet
    """
    assert type(df) is DataFrame
    assert type(path) is str
    assert type(label) is str
    column_indices: Dict[str, int] = {}
    negative_columns: Dict[str, bool] = {}
    column_names: OrderedDict[str, List[str]] = OrderedDict(
        {
            "frequency": ["frequency", "freq", "f"],
            "imaginary": ['z"', "z''", "z_im", "zim", "imaginary", "imag", "im"],
            "real": ["z'", "z_re", "zre", "real", "re"],
            "magnitude": ["|z|", "z", "magnitude", "modulus", "mag", "mod"],
            "phase": ["phase", "phz", "phi"],
        }
    )
    i: int
    col: str
    for i, col in enumerate(df.columns):
        col = col.lower()
        key: str
        alternatives: List[str]
        for key, alternatives in column_names.items():
            if key in column_indices:
                # Skip columns that have already been identified
                continue
            alt: str
            for alt in alternatives:
                if col.startswith(alt) or col.startswith(f"-{alt}"):
                    column_indices[key] = i
                    negative_columns[key] = col.startswith("-")
                    break
            if key in column_indices:
                # The column has been identified
                break
    assert len(column_indices) >= 3, column_indices
    assert "frequency" in column_indices
    frequency: List[float] = []
    real: List[float] = []
    imaginary: List[float] = []
    magnitude: List[float] = []
    phase: Union[List[float], ndarray] = []
    for row in df.values:
        frequency.append(row[column_indices["frequency"]])
        if "real" in column_indices and "imaginary" in column_indices:
            re: float = row[column_indices["real"]]
            if negative_columns["real"]:
                re *= -1
            real.append(re)
            im: float = row[column_indices["imaginary"]]
            if negative_columns["imaginary"]:
                im *= -1
            imaginary.append(im)
        elif "magnitude" in column_indices and "phase" in column_indices:
            mag: float = row[column_indices["magnitude"]]
            magnitude.append(mag)
            phi: float = row[column_indices["phase"]]
            if negative_columns["phase"]:
                phi *= -1
            phase.append(phi)  # type: ignore
        else:
            raise Exception("Unsupported file format/structure: {path}")
    if len(phase) > 0:
        assert len(phase) == len(magnitude)
        phase = array(phase)
        if max(abs(phase)) > pi:
            phase = radians(phase)
        for mag, phi in zip(magnitude, phase):
            Z: complex = cmath.rect(mag, phi)
            real.append(Z.real)
            imaginary.append(Z.imag)
    assert len(frequency) == len(real) == len(imaginary) > 0, (
        len(frequency),
        len(real),
        len(imaginary),
    )
    return DataSet(
        array(frequency),
        array(list(map(lambda _: complex(*_), zip(real, imaginary)))),
        path=path,
        label=label,
    )



# pyimpspec is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 pyimpspec developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of pyimpspec's dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from os.path import (
    basename,
    exists,
    splitext,
)
from typing import (
    Dict,
    IO,
    List,
    Tuple,
)
from numpy import (
    integer,
    issubdtype,
)
from pyimpspec.data.data_set import (
    DataSet,
    dataframe_to_dataset,
)
from pandas import DataFrame
from string import printable


def _extract_primary_data(
    num_freq: int, lines: List[str]
) -> Tuple[List[float], List[float], List[float]]:
    assert issubdtype(type(num_freq), integer), num_freq
    assert type(lines) is list and all(map(lambda _: type(_) is str, lines))
    assert len(lines) > num_freq
    re: List[float] = []
    im: List[float] = []
    f: List[float] = []
    while lines:
        line: str = lines.pop(0)
        values: List[str] = line.replace(",", ".").split()
        assert len(values) == 3, values
        re.append(float(values.pop(0)))
        im.append(float(values.pop(0)))
        f.append(float(values.pop(0)))
        if len(re) == num_freq:
            break
    return (
        re,
        im,
        f,
    )


def parse_ids(path: str) -> List[DataSet]:
    """
    Parse an Ivium .ids or .idf file containing one or more impedance spectra.

    Parameters
    ----------
    path: str
        The path to the file to process.

    Returns
    -------
    DataSet
    """
    assert type(path) is str and exists(path)
    default_label: str = splitext(basename(path))[0]
    fp: IO
    with open(path, "r", encoding="latin1") as fp:
        content: str = "".join(map(lambda _: _ if _ in printable else "", fp.read()))
        lines: List[str] = list(
            filter(lambda _: _ != "", map(str.strip, content.split("\n")))
        )
    raw_data_sets: Dict[str, dict] = {}
    while lines:
        while lines:
            line: str = lines.pop(0)
            if line.startswith("Title="):
                break
        if len(lines) == 0:
            break
        label: str = line[6:].strip()
        while lines:
            line = lines.pop(0)
            if "primary_data" in line:
                break
        assert len(lines) > 0, f"Failed to find the primary data in '{path}'"
        assert int(lines.pop(0)) == 3  # Number of columns
        num_freq: int = int(lines.pop(0))
        re: List[float]
        im: List[float]
        f: List[float]
        re, im, f = _extract_primary_data(num_freq, lines)
        assert len(re) == len(im) == len(f) == num_freq
        if label == "":
            label = default_label
        if label in raw_data_sets:
            i: int = 2
            while f"{label} ({i})" in raw_data_sets:
                i += 1
            label = f"{label} ({i})"
        raw_data_sets[label] = {
            "freq": f,
            "real": re,
            "imag": im,
        }
    assert len(raw_data_sets) > 0, raw_data_sets
    assert len(lines) == 0, lines
    data_sets: List[DataSet] = []
    for label, values in raw_data_sets.items():
        data_sets.append(dataframe_to_dataset(DataFrame.from_dict(values), path, label))
    assert len(data_sets) == len(raw_data_sets), (
        len(data_sets),
        len(raw_data_sets),
    )
    return data_sets



# pyimpspec is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 pyimpspec developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of pyimpspec's dependencies and/or sources of portions of code are included in
# the LICENSES folder.

# This module implements the Bayesian Hilbert transform method
# pyDRTtools commit: 65ea54d9332a0c6594de852f0242a88e20ec4427

from contextlib import redirect_stdout
from multiprocessing import (
    Pool,
    cpu_count,
)
from os import devnull
from typing import (
    Callable,
    IO,
    List,
    Optional,
    Tuple,
)
from numpy import (
    array,
    diag,
    empty_like,
    exp,
    eye,
    floating,
    integer,
    issubdtype,
    log as ln,
    log10 as log,
    logical_and,
    logspace,
    ndarray,
    pi,
    sqrt,
    squeeze,
    sum as array_sum,
    zeros,
)
from numpy.linalg import (
    cholesky,
    inv as invert,
    norm,
    solve as solve_linalg,
)
from numpy.random import (
    rand,
)
from scipy.optimize import (
    OptimizeResult,
    minimize,
)
from scipy.stats import multivariate_normal
from pyimpspec.data import DataSet
from pyimpspec.analysis.fitting import _calculate_residuals
from .result import (
    DRTError,
    DRTResult,
    _calculate_chisqr,
)
from .tr_rbf import (
    _x_to_gamma,
    _compute_epsilon,
    RBF_TYPES,
    RBF_SHAPES,
)
import pyimpspec.progress as progress


def _compute_res_scores(res: ndarray, band: ndarray) -> ndarray:
    # Count the points fallen inside the 1, 2, and 3 sigma credible bands
    count: ndarray = zeros(3)
    i: int
    for i in range(3):
        count[i] = array_sum(logical_and(res < (i + 1) * band, res > -(i + 1) * band))
    return count / len(res)


def _compute_SHD(
    mu_P: ndarray,
    Sigma_P: ndarray,
    mu_Q: ndarray,
    Sigma_Q: ndarray,
) -> ndarray:
    # Squared Hellinger distance
    sigma_P: ndarray = sqrt(diag(Sigma_P))
    sigma_Q: ndarray = sqrt(diag(Sigma_Q))
    sum_cov: ndarray = sigma_P**2 + sigma_Q**2
    prod_cov: ndarray = sigma_P * sigma_Q
    return 1.0 - sqrt(2.0 * prod_cov / sum_cov) * exp(
        -0.25 * (mu_P - mu_Q) ** 2 / sum_cov
    )


def _compute_JSD(
    mu_P: ndarray,
    Sigma_P: ndarray,
    mu_Q: ndarray,
    Sigma_Q: ndarray,
    num_samples: int,
) -> ndarray:
    # Jensen-Shannon distance (JSD)
    JSD: ndarray = empty_like(mu_P)
    i: int
    for i in range(mu_P.size):
        RV_p = multivariate_normal(mean=mu_P[i], cov=Sigma_P[i, i])
        RV_q = multivariate_normal(mean=mu_Q[i], cov=Sigma_Q[i, i])
        x: ndarray = RV_p.rvs(num_samples)
        p_x: ndarray = RV_p.pdf(x)
        q_x: ndarray = RV_q.pdf(x)
        m_x: ndarray = (p_x + q_x) / 2.0
        y: ndarray = RV_q.rvs(num_samples)
        p_y: ndarray = RV_p.pdf(y)
        q_y: ndarray = RV_q.pdf(y)
        m_y: ndarray = (p_y + q_y) / 2.0
        dKL_pm: ndarray = ln(p_x / m_x).mean()
        dKL_qm: ndarray = ln(q_y / m_y).mean()
        JSD[i] = 0.5 * (dKL_pm + dKL_qm)
    return JSD


def _NMLL_fct(
    theta: ndarray,
    Z: ndarray,
    A: ndarray,
    L: ndarray,
    num_freqs: int,
    num_taus: int,
) -> ndarray:
    sigma_n: ndarray
    sigma_beta: ndarray
    sigma_lambda: ndarray
    sigma_n, sigma_beta, sigma_lambda = theta
    W: ndarray = (
        1 / (sigma_beta**2) * eye(num_taus + 1) + 1 / (sigma_lambda**2) * L.T @ L
    )
    # W = 0.5 * (W.T + W)
    K_agm: ndarray = 1 / (sigma_n**2) * (A.T @ A) + W
    # K_agm = 0.5 * (K_agm.T + K_agm)
    L_W: ndarray = cholesky(W)
    L_agm: ndarray = cholesky(K_agm)
    # Compute mu_x
    u: ndarray = solve_linalg(L_agm.T, solve_linalg(L_agm, A.T @ Z))
    mu_x: ndarray = 1 / (sigma_n**2) * u
    # Compute loss
    E_mu_x: ndarray = 0.5 / (sigma_n**2) * norm(A @ mu_x - Z) ** 2 + 0.5 * (
        mu_x.T @ (W @ mu_x)
    )
    val_1: ndarray = array_sum(ln(diag(L_W)))
    val_2: ndarray = -array_sum(ln(diag(L_agm)))
    val_3: ndarray = -num_freqs / 2.0 * ln(sigma_n**2)
    val_4: ndarray = -E_mu_x
    val_5: ndarray = -num_freqs / 2 * ln(2 * pi)
    return -(val_1 + val_2 + val_3 + val_4 + val_5)


def _compute_A_re(w: ndarray, tau: ndarray) -> ndarray:
    num_freqs: int = w.shape[0]
    num_taus: int = tau.shape[0]
    A_re: ndarray = zeros(
        (
            num_freqs,
            num_taus + 1,
        )
    )
    A_re[:, 0] = 1.0
    i: int
    j: int
    for i in range(0, num_freqs):
        for j in range(0, num_taus):
            A_re[i, j + 1] = (
                0.5
                / (1 + (w[i] * tau[j]) ** 2)
                * ln(
                    (tau[j] if j == num_taus - 1 else tau[j + 1])
                    / (tau[j] if j == 0 else tau[j - 1])
                )
            )
    return A_re


def _compute_A_H_re(w: ndarray, tau: ndarray) -> ndarray:
    return _compute_A_re(w, tau)[:, 1:]


def _compute_A_im(w: ndarray, tau: ndarray) -> ndarray:
    num_freqs: int = w.shape[0]
    num_taus: int = tau.shape[0]
    A_im: ndarray = zeros(
        (
            num_freqs,
            num_taus + 1,
        )
    )
    A_im[:, 0] = w
    i: int
    j: int
    for i in range(0, num_freqs):
        for j in range(0, num_taus):
            A_im[i, j + 1] = (
                -0.5
                * (w[i] * tau[j])
                / (1 + (w[i] * tau[j]) ** 2)
                * ln(
                    (tau[j] if j == num_taus - 1 else tau[j + 1])
                    / (tau[j] if j == 0 else tau[j - 1])
                )
            )
    return A_im


def _compute_A_H_im(w: ndarray, tau: ndarray) -> ndarray:
    return _compute_A_im(w, tau)[:, 1:]


def _compute_L(tau: ndarray, derivative_order: int) -> ndarray:
    num_taus: int = tau.shape[0]
    L: ndarray = zeros(
        (
            num_taus - 2,
            num_taus + 1,
        )
    )
    i: int
    delta_loc: float
    if derivative_order == 1:
        for i in range(0, num_taus - 2):
            delta_loc = ln(tau[i + 1] / tau[i])
            factors: ndarray = array([1.0, -2.0, 1.0])
            if i == 0 or i == num_taus - 3:
                factors *= 2
            L[i, i + 1] = factors[0] / (delta_loc**2)
            L[i, i + 2] = factors[1] / (delta_loc**2)
            L[i, i + 3] = factors[2] / (delta_loc**2)
    elif derivative_order == 2:
        for i in range(0, num_taus - 2):
            delta_loc = ln(tau[i + 1] / tau[i])
            if i == 0:
                L[i, i + 1] = -3.0 / (2 * delta_loc)
                L[i, i + 2] = 4.0 / (2 * delta_loc)
                L[i, i + 3] = -1.0 / (2 * delta_loc)
            elif i == num_taus - 2:
                L[i, i] = 1.0 / (2 * delta_loc)
                L[i, i + 1] = -4.0 / (2 * delta_loc)
                L[i, i + 2] = 3.0 / (2 * delta_loc)
            else:
                L[i, i] = 1.0 / (2 * delta_loc)
                L[i, i + 2] = -1.0 / (2 * delta_loc)
    else:
        raise Exception(f"Unsupported derivative order: {derivative_order}")
    return L


def _calculate_scores(
    theta_0: ndarray,
    f: ndarray,
    Z_exp: ndarray,
    out_dict_real: dict,
    out_dict_imag: dict,
    num_samples: int,
) -> dict:
    progress_message: str = "Calculating scores"
    num_scores: int = 12
    progress.update_every_N_percent(0, total=num_scores, message=progress_message)
    # scores
    # s_mu - distance between means:
    mu_Z_DRT_re: ndarray = out_dict_real["mu_Z_DRT"]
    mu_Z_H_re: ndarray = out_dict_imag["mu_Z_H"]
    discrepancy_re = norm(mu_Z_DRT_re - mu_Z_H_re) / (
        norm(mu_Z_DRT_re) + norm(mu_Z_H_re)
    )
    s_mu_re: float = 1.0 - discrepancy_re  # type: ignore
    progress.update_every_N_percent(1, total=num_scores, message=progress_message)
    mu_Z_DRT_im: ndarray = out_dict_imag["mu_Z_DRT"]
    mu_Z_H_im: ndarray = out_dict_real["mu_Z_H"]
    discrepancy_im = norm(mu_Z_DRT_im - mu_Z_H_im) / (
        norm(mu_Z_DRT_im) + norm(mu_Z_H_im)
    )
    s_mu_im: float = 1.0 - discrepancy_im  # type: ignore
    progress.update_every_N_percent(2, total=num_scores, message=progress_message)
    # s_JSD - Jensen-Shannon Distance:
    # we need the means (above) and covariances (below)
    # for the computation of the JSD
    Sigma_Z_DRT_re: ndarray = out_dict_real["Sigma_Z_DRT"]
    Sigma_Z_DRT_im: ndarray = out_dict_imag["Sigma_Z_DRT"]
    Sigma_Z_H_re: ndarray = out_dict_imag["Sigma_Z_H"]
    Sigma_Z_H_im: ndarray = out_dict_real["Sigma_Z_H"]
    # s_res - residual score:
    # real part
    # retrieve distribution of R_inf
    mu_R_inf: float = out_dict_real["mu_gamma"][0]
    cov_R_inf: ndarray = diag(out_dict_real["Sigma_gamma"])[0]
    # we will also need omega an estimate of the error
    sigma_n_im: ndarray = out_dict_imag["theta"][0]
    # R_inf+Z_H_re-Z_exp has:
    # mean:
    res_re: ndarray = mu_R_inf + mu_Z_H_re - Z_exp.real
    # std:
    band_re: ndarray = sqrt(cov_R_inf + diag(Sigma_Z_H_re) + sigma_n_im**2)
    s_res_re: ndarray = _compute_res_scores(res_re, band_re)
    progress.update_every_N_percent(5, total=num_scores, message=progress_message)
    # imaginary part
    # retrieve distribution of L_0
    mu_L_0: float = out_dict_imag["mu_gamma"][0]
    cov_L_0: ndarray = diag(out_dict_imag["Sigma_gamma"])[0]
    # we will also need omega
    omega_vec: ndarray = 2.0 * pi * f
    # and an estimate of the error
    sigma_n_re: ndarray = out_dict_real["theta"][0]
    # R_inf+Z_H_re-Z_exp has:
    # mean:
    res_im: ndarray = omega_vec * mu_L_0 + mu_Z_H_im - Z_exp.imag
    # std:
    band_im: ndarray = sqrt(
        (omega_vec**2) * cov_L_0 + diag(Sigma_Z_H_im) + sigma_n_re**2
    )
    s_res_im: ndarray = _compute_res_scores(res_im, band_im)
    progress.update_every_N_percent(8, total=num_scores, message=progress_message)
    # Squared Hellinger distance (SHD)
    # which is bounded between 0 and 1
    # we are going to score w.r.t. the Hellinger distance (HD)
    # the score uses 1 to mean good (this means close)
    # and 0 means bad (far away) => that's the opposite of the distance
    SHD_re: ndarray = _compute_SHD(mu_Z_DRT_re, Sigma_Z_DRT_re, mu_Z_H_re, Sigma_Z_H_re)
    s_HD_re: float = 1.0 - sqrt(SHD_re).mean()
    progress.update_every_N_percent(9, total=num_scores, message=progress_message)
    SHD_im: ndarray = _compute_SHD(mu_Z_DRT_im, Sigma_Z_DRT_im, mu_Z_H_im, Sigma_Z_H_im)
    s_HD_im: float = 1.0 - sqrt(SHD_im).mean()
    progress.update_every_N_percent(10, total=num_scores, message=progress_message)
    # compute the Jensen-Shannon distance (JSD)
    # the JSD is a symmetrized relative entropy (discrepancy), so highest value means more entropy
    # we are going to reverse that by taking (ln(2)-JSD)/ln(2)
    # which means higher value less relative entropy (discrepancy)
    JSD_re: ndarray = _compute_JSD(
        mu_Z_DRT_re,
        Sigma_Z_DRT_re,
        mu_Z_H_re,
        Sigma_Z_H_re,
        num_samples,
    )
    s_JSD_re: float = (ln(2) - JSD_re.mean()) / ln(2)
    progress.update_every_N_percent(11, total=num_scores, message=progress_message)
    JSD_im: ndarray = _compute_JSD(
        mu_Z_DRT_im,
        Sigma_Z_DRT_im,
        mu_Z_H_im,
        Sigma_Z_H_im,
        num_samples,
    )
    s_JSD_im: float = (ln(2) - JSD_im.mean()) / ln(2)
    progress.update_every_N_percent(12, total=num_scores, message=progress_message)
    return {
        "hellinger_distance": complex(s_HD_re, s_HD_im),
        "jensen_shannon_distance": complex(s_JSD_re, s_JSD_im),
        "mean": complex(s_mu_re, s_mu_im),
        "residuals_1sigma": complex(s_res_re[0], s_res_im[0]),
        "residuals_2sigma": complex(s_res_re[1], s_res_im[1]),
        "residuals_3sigma": complex(s_res_re[2], s_res_im[2]),
    }


def _single_hilbert_transform_estimate(
    theta_0: ndarray,
    Z_exp: ndarray,
    A: ndarray,
    A_H: ndarray,
    L: ndarray,
    num_freqs: int,
    num_taus: int,
):
    fp: IO
    with open(devnull, "w") as fp:
        with redirect_stdout(fp):
            res: OptimizeResult = minimize(
                _NMLL_fct,
                squeeze(theta_0),
                args=(Z_exp, A, L, num_freqs, num_taus),
                options={"gtol": 1e-8, "disp": True},
            )
    sigma_n: float
    sigma_beta: float
    sigma_lambda: float
    sigma_n, sigma_beta, sigma_lambda = res.x
    # Compute the probability density functions of data regression
    # $K_agm = A.T A +\lambda L.T L$
    W: ndarray = (
        1 / (sigma_beta**2) * eye(num_taus + 1) + 1 / (sigma_lambda**2) * L.T @ L
    )
    K_agm: ndarray = 1 / (sigma_n**2) * (A.T @ A) + W
    # Cholesky factorization
    L_agm: ndarray = cholesky(K_agm)
    inv_L_agm: ndarray = invert(L_agm)
    inv_K_agm: ndarray = inv_L_agm.T @ inv_L_agm
    # Compute the gamma ~ N(mu_gamma, Sigma_gamma)
    Sigma_gamma: ndarray = inv_K_agm
    mu_gamma: ndarray = 1 / (sigma_n**2) * (Sigma_gamma @ A.T) @ Z_exp.real
    # Compute, from gamma, the Z ~ N(mu_Z, Sigma_Z)
    mu_Z: ndarray = A @ mu_gamma
    Sigma_Z: ndarray = A @ (Sigma_gamma @ A.T) + sigma_n**2 * eye(num_freqs)
    # Compute, from gamma, the Z_DRT ~ N(mu_Z_DRT, Sigma_Z_DRT)
    A_DRT: ndarray = A[:, 1:]
    mu_gamma_DRT: ndarray = mu_gamma[1:]
    Sigma_gamma_DRT: ndarray = Sigma_gamma[1:, 1:]
    mu_Z_DRT: ndarray = A_DRT @ mu_gamma_DRT
    Sigma_Z_DRT: ndarray = A_DRT @ (Sigma_gamma_DRT @ A_DRT.T)
    # Compute, from gamma, the Z_H_conj ~ N(mu_Z_H_conj, Sigma_Z_H_conj)
    mu_Z_H: ndarray = A_H @ mu_gamma[1:]
    Sigma_Z_H: ndarray = A_H @ (Sigma_gamma[1:, 1:] @ A_H.T)
    return {
        "mu_gamma": mu_gamma,
        "Sigma_gamma": Sigma_gamma,
        "mu_Z": mu_Z,
        "Sigma_Z": Sigma_Z,
        "mu_Z_DRT": mu_Z_DRT,
        "Sigma_Z_DRT": Sigma_Z_DRT,
        "mu_Z_H": mu_Z_H,
        "Sigma_Z_H": Sigma_Z_H,
        "theta": res.x,
    }


def _calculate_symmetry_score(
    result: Tuple[float, ndarray, dict, dict],
    tau_fine: ndarray,
    tau: ndarray,
    epsilon: float,
    rbf_type: str,
) -> float:
    # Calculate gamma for each result and check if it is asymmetrical
    chisqr: float
    theta_0: ndarray
    data_real: dict
    data_imag: dict
    chisqr, theta_0, data_real, data_imag = result
    _, gamma = _x_to_gamma(
        data_real["mu_gamma"][1:],
        tau_fine,
        tau,
        epsilon,
        rbf_type,
    )
    min_gamma: float = abs(min(gamma))
    max_gamma: float = abs(max(gamma))
    score: float = 1.0 - ((max_gamma - min_gamma) / max(min_gamma, max_gamma))
    _, gamma = _x_to_gamma(
        data_imag["mu_gamma"][1:],
        tau_fine,
        tau,
        epsilon,
        rbf_type,
    )
    min_gamma = abs(min(gamma))
    max_gamma = abs(max(gamma))
    score += 1.0 - ((max_gamma - min_gamma) / max(min_gamma, max_gamma))
    return score / 2.0


def _hilbert_transform_process(
    args: tuple,
) -> Optional[Tuple[float, ndarray, dict, dict]]:
    theta_0: ndarray
    w: ndarray
    Z: ndarray
    A_re: ndarray
    A_im: ndarray
    A_H_re: ndarray
    A_H_im: ndarray
    L: ndarray
    num_freqs: int
    num_taus: int
    maximum_symmetry: float
    tau_fine: ndarray
    tau: ndarray
    epsilon: float
    rbf_type: str
    (
        theta_0,
        w,
        Z,
        A_re,
        A_im,
        A_H_re,
        A_H_im,
        L,
        num_freqs,
        num_taus,
        maximum_symmetry,
        tau_fine,
        tau,
        epsilon,
        rbf_type,
    ) = args
    try:
        data_real: dict = _single_hilbert_transform_estimate(
            theta_0,
            Z.real,
            A_re,
            A_H_im,
            L,
            num_freqs,
            num_taus,
        )
        theta_0 = data_real["theta"]
        data_imag: dict = _single_hilbert_transform_estimate(
            theta_0,
            Z.imag,
            A_im,
            A_H_re,
            L,
            num_freqs,
            num_taus,
        )
        mu_R_inf: float = data_real["mu_gamma"][0]
        mu_Z_H_re: ndarray = data_imag["mu_Z_H"]
        mu_L_0: float = data_imag["mu_gamma"][0]
        mu_Z_H_im: ndarray = data_real["mu_Z_H"]
        Z_fit: ndarray = array(
            list(
                map(
                    lambda _: complex(*_),
                    zip(
                        mu_R_inf + mu_Z_H_re,
                        w * mu_L_0 + mu_Z_H_im,
                    ),
                )
            )
        )
        z: complex = array_sum((Z_fit - Z) ** 2 / Z)
        chisqr: float = sqrt(z.real**2 + z.imag**2)
        result: Tuple[float, ndarray, dict, dict] = (
            chisqr,
            theta_0,
            data_real,
            data_imag,
        )
        if (
            _calculate_symmetry_score(
                result,
                tau_fine,
                tau,
                epsilon,
                rbf_type,
            )
            > maximum_symmetry
        ):
            # The result is most likely poor (lots of strong oscillation).
            return None
        return (
            chisqr,
            theta_0,
            data_real,
            data_imag,
        )
    except Exception:
        return None


def _calculate_drt_bht(
    data: DataSet,
    rbf_type: str,
    derivative_order: int,
    rbf_shape: str,
    shape_coeff: float,
    num_samples: int,
    num_attempts: int,
    maximum_symmetry: float,
    num_procs: int,
) -> DRTResult:
    assert (
        hasattr(data, "get_frequency")
        and callable(data.get_frequency)
        and hasattr(data, "get_impedance")
        and callable(data.get_impedance)
    ), data
    assert type(rbf_type) is str and rbf_type in RBF_TYPES, rbf_type
    assert (
        issubdtype(type(derivative_order), integer) and 1 <= derivative_order <= 2
    ), derivative_order
    assert type(rbf_shape) is str and rbf_shape in RBF_SHAPES, rbf_shape
    assert (
        issubdtype(
            type(shape_coeff),
            floating,
        )
        and shape_coeff > 0.0
    ), shape_coeff
    assert issubdtype(type(num_samples), integer) and num_samples > 0, num_samples
    assert issubdtype(type(num_attempts), integer) and num_attempts >= 1, num_attempts
    assert (
        issubdtype(
            type(maximum_symmetry),
            floating,
        )
        and 0.0 <= maximum_symmetry <= 1.0
    ), maximum_symmetry
    assert issubdtype(type(num_procs), integer), num_procs
    if num_procs < 1:
        num_procs = cpu_count()
    f: ndarray = data.get_frequency()
    Z: ndarray = data.get_impedance()
    tau: ndarray = 1 / f
    tau_fine: ndarray = logspace(
        log(tau.min()) - 0.5,
        log(tau.max()) + 0.5,
        10 * f.shape[0],
    )
    w: ndarray = 2 * pi * f
    num_freqs: int = f.shape[0]
    num_taus: int = tau.shape[0]
    epsilon: float = _compute_epsilon(
        f,
        rbf_shape,
        shape_coeff,
        rbf_type,
    )
    A_re: ndarray = _compute_A_re(w, tau)
    A_im: ndarray = _compute_A_im(w, tau)
    A_H_re: ndarray = _compute_A_H_re(w, tau)
    A_H_im: ndarray = _compute_A_H_im(w, tau)
    L: ndarray = _compute_L(tau, derivative_order)
    theta_0_generator: Callable = lambda: 10 ** (6 * rand(3, 1) - 3)
    with Pool(num_procs) as pool:
        args = (
            (
                theta_0_generator(),
                w,
                Z,
                A_re,
                A_im,
                A_H_re,
                A_H_im,
                L,
                num_freqs,
                num_taus,
                maximum_symmetry,
                tau_fine,
                tau,
                epsilon,
                rbf_type,
            )
            for _ in range(0, num_attempts)
        )
        results: List[Tuple[float, ndarray, dict, dict]] = []
        progress_message: str = "Calculating Hilbert transforms"
        progress.update_every_N_percent(0, total=num_attempts, message=progress_message)
        for i, res in enumerate(
            pool.imap_unordered(
                _hilbert_transform_process,
                args,
            )
        ):
            progress.update_every_N_percent(
                i + 1,
                total=num_attempts,
                message=progress_message,
            )
            if res is not None:
                results.append(res)
    if len(results) == 0:
        raise DRTError("Failed to perform calculations! Try tweaking the settings.")
    results.sort(key=lambda _: _[0])
    theta_0: ndarray
    data_real: dict
    data_imag: dict
    _, theta_0, data_real, data_imag = results[0]
    # Scores seem to be fine based on comparison with the ZARC example used in the article
    scores: dict = _calculate_scores(
        theta_0,
        f,
        Z,
        data_real,
        data_imag,
        num_samples,
    )
    # Real part
    mu_Z_re: ndarray = data_real["mu_Z"]
    cov_Z_r: ndarray = diag(data_real["Sigma_Z"])
    mu_R_inf: float = data_real["mu_gamma"][0]
    cov_R_inf: ndarray = diag(data_real["Sigma_gamma"])[0]
    mu_Z_DRT_re: ndarray = data_real["mu_Z_DRT"]
    cov_Z_DRT_re: ndarray = diag(data_real["Sigma_Z_DRT"])
    mu_Z_H_im: ndarray = data_real["mu_Z_H"]
    cov_Z_H_im: ndarray = diag(data_real["Sigma_Z_H"])
    sigma_n_re = data_real["theta"][0]
    # Imaginary part
    mu_Z_im: ndarray = data_imag["mu_Z"]
    cov_Z_im: ndarray = diag(data_imag["Sigma_Z"])
    mu_L_0: float = data_imag["mu_gamma"][0]
    cov_L_0: ndarray = diag(data_imag["Sigma_gamma"])[0]
    mu_Z_DRT_im: ndarray = data_imag["mu_Z_DRT"]
    cov_Z_DRT_im: ndarray = diag(data_imag["Sigma_Z_DRT"])
    mu_Z_H_re: ndarray = data_imag["mu_Z_H"]
    cov_Z_H_re: ndarray = diag(data_imag["Sigma_Z_H"])
    sigma_n_im: float = data_imag["theta"][0]
    # Means and bounds
    mu_Z_H_re_agm: ndarray = mu_R_inf + mu_Z_H_re
    mu_Z_H_im_agm: ndarray = w * mu_L_0 + mu_Z_H_im
    # band_agm_re = sqrt(cov_R_inf + cov_Z_H_re + sigma_n_im**2)
    # band_agm_im = sqrt((w**2) * cov_L_0 + cov_Z_H_im + sigma_n_re**2)
    Z_fit: ndarray = array(
        list(
            map(
                lambda _: complex(*_),
                zip(
                    mu_Z_H_re_agm,
                    mu_Z_H_im_agm,
                ),
            )
        )
    )
    return DRTResult(
        "BHT",
        *_x_to_gamma(
            data_real["mu_gamma"][1:],
            tau_fine,
            tau,
            epsilon,
            rbf_type,
        ),
        f,
        Z_fit,
        *_calculate_residuals(Z, Z_fit),
        # "tr-rbf" method with credible_intervals
        array([]),  # Mean
        array([]),  # Lower bound
        array([]),  # Upper bound
        # "bht" method
        _x_to_gamma(
            data_imag["mu_gamma"][1:],
            tau_fine,
            tau,
            epsilon,
            rbf_type,
        )[1],
        scores,
        # Stats
        _calculate_chisqr(Z, Z_fit),
        -1.0,
    )



# pyimpspec is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 pyimpspec developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of pyimpspec's dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from dataclasses import dataclass
from multiprocessing import (
    Pool,
    Value,
    cpu_count,
)
from typing import (
    Callable,
    List,
    Tuple,
    Optional,
)
from traceback import format_exc
from lmfit import (
    Parameters,
    minimize,
)
from lmfit.minimizer import MinimizerResult
from numpy import (
    abs,
    angle,
    array,
    float64,
    floating,
    inf,
    integer,
    issubdtype,
    log10 as log,
    min,
    max,
    ndarray,
    pi,
    sum as array_sum,
    zeros,
)
from numpy.linalg import (
    inv,
    pinv,
)
from pyimpspec.analysis.fitting import (
    _from_lmfit,
    _to_lmfit,
    _interpolate,
    _calculate_residuals,
    FittingError,
)
from pyimpspec.circuit import parse_cdc
from pyimpspec.circuit.base import Connection
from pyimpspec.circuit.circuit import Circuit
from pyimpspec.data import DataSet
import pyimpspec.progress as progress


@dataclass(frozen=True)
class TestResult:
    """
    An object representing the results of a linear Kramers-Kronig test applied to a data set.

    Parameters
    ----------
    circuit: Circuit
        The fitted circuit.

    num_RC: int
        The final number of RC elements in the fitted model (Boukamp, 1995).

    mu: float
        The mu-value of the final fit (eq. 21 in Schönleber et al., 2014).

    pseudo_chisqr: float
        The pseudo chi-squared fit value (eq. 14 in Boukamp, 1995).

    frequency: ndarray
        The frequencies used to perform the test.

    impedance: ndarray
        The impedance produced by the fitted circuit at each of the tested frequencies.

    real_residual: ndarray
        The residuals for the real parts (eq. 15 in Schönleber et al., 2014).

    imaginary_residual: ndarray
        The residuals for the imaginary parts (eq. 16 in Schönleber et al., 2014).
    """

    circuit: Circuit
    num_RC: int
    mu: float
    pseudo_chisqr: float
    frequency: ndarray
    impedance: ndarray
    real_residual: ndarray
    imaginary_residual: ndarray

    def __repr__(self) -> str:
        return f"TestResult (num_RC={self.num_RC}, {hex(id(self))})"

    def get_frequency(self, num_per_decade: int = -1) -> ndarray:
        assert issubdtype(type(num_per_decade), integer), num_per_decade
        if num_per_decade > 0:
            return _interpolate(self.frequency, num_per_decade)
        return self.frequency

    def get_impedance(self, num_per_decade: int = -1) -> ndarray:
        assert issubdtype(type(num_per_decade), integer), num_per_decade
        if num_per_decade > 0:
            return self.circuit.impedances(self.get_frequency(num_per_decade))
        return self.impedance

    def get_nyquist_data(self, num_per_decade: int = -1) -> Tuple[ndarray, ndarray]:
        """
        Get the data necessary to plot this TestResult as a Nyquist plot: the real and the negative imaginary parts of the impedances.

        Parameters
        ----------
        num_per_decade: int = -1

        Returns
        -------
        Tuple[ndarray, ndarray]
        """
        assert issubdtype(type(num_per_decade), integer), num_per_decade
        if num_per_decade > 0:
            Z: ndarray = self.get_impedance(num_per_decade)
            return (
                Z.real,
                -Z.imag,
            )
        return (
            self.impedance.real,
            -self.impedance.imag,
        )

    def get_bode_data(
        self, num_per_decade: int = -1
    ) -> Tuple[ndarray, ndarray, ndarray]:
        """
        Get the data necessary to plot this TestResult as a Bode plot: the frequencies, the absolute magnitudes of the impedances, and the negative phase angles/shifts of the impedances in degrees.

        Parameters
        ----------
        num_per_decade: int = -1

        Returns
        -------
        Tuple[ndarray, ndarray, ndarray]
        """
        assert issubdtype(type(num_per_decade), integer), num_per_decade
        if num_per_decade > 0:
            freq: ndarray = self.get_frequency(num_per_decade)
            Z: ndarray = self.circuit.impedances(freq)
            return (
                freq,
                abs(Z),
                -angle(Z, deg=True),
            )
        return (
            self.frequency,
            abs(self.impedance),
            -angle(self.impedance, deg=True),
        )

    def get_residual_data(self) -> Tuple[ndarray, ndarray, ndarray]:
        """
        Get the data necessary to plot the relative residuals for this TestResult: the frequencies, the relative residuals for the real parts of the impedances in percents, and the relative residuals for the imaginary parts of the impedances in percents.

        Returns
        -------
        Tuple[ndarray, ndarray, ndarray]
        """
        return (
            self.frequency,
            self.real_residual * 100,
            self.imaginary_residual * 100,
        )

    def calculate_score(self, mu_criterion: float) -> float:
        """
        Calculate a score based on the provided mu-criterion and the statistics of the result.
        A result with a mu-value greater than or equal to the mu-criterion will get a score of -numpy.inf.

        Parameters
        ----------
        mu_criterion: float
            The mu-criterion to apply.
            See `perform_test` for details.

        Returns
        -------
        float
        """
        return (
            -inf
            if self.mu >= mu_criterion
            else -log(self.pseudo_chisqr) / (abs(mu_criterion - self.mu) ** 0.75)
        )


def _calculate_tau(i: int, num_RC: int, tau_min: float64, tau_max: float64) -> float:
    # Calculate time constants according to eq. 12 in Schönleber et al. (2014)
    assert issubdtype(type(i), integer), i
    assert issubdtype(type(num_RC), integer), num_RC
    assert issubdtype(type(tau_min), floating), tau_min
    assert issubdtype(type(tau_max), floating), tau_max
    return pow(10, (log(tau_min) + (i - 1) / (num_RC - 1) * log(tau_max / tau_min)))


def _generate_time_constants(w: ndarray, num_RC: int) -> ndarray:
    assert type(w) is ndarray, w
    assert issubdtype(type(num_RC), integer), num_RC
    taus: ndarray = zeros(shape=(num_RC,))
    tau_min: float64 = 1 / max(w)
    tau_max: float64 = 1 / min(w)
    taus[0] = tau_min
    taus[-1] = tau_max
    i: int
    if num_RC > 1:
        for i in range(2, num_RC):
            taus[i - 1] = _calculate_tau(i, num_RC, tau_min, tau_max)
    return taus


def _generate_weight(Z_exp: ndarray) -> ndarray:
    assert type(Z_exp) is ndarray, Z_exp
    # See eq. 13 in Boukamp (1995)
    return (Z_exp.real**2 + Z_exp.imag**2) ** -1


def _calculate_pseudo_chisqr(Z_exp: ndarray, Z_fit: ndarray, weight: ndarray) -> float:
    assert type(Z_exp) is ndarray, Z_exp
    assert type(Z_fit) is ndarray, Z_fit
    assert type(weight) is ndarray, weight
    # See eq. 14 in Boukamp (1995)
    return float(
        array_sum(
            weight * ((Z_exp.real - Z_fit.real) ** 2 + (Z_exp.imag - Z_fit.imag) ** 2)
        )
    )


def _calculate_mu(params: Parameters) -> float:
    assert type(params) is Parameters, params
    # Calculates the mu-value based on the fitted parameters according to eq. 21 in
    # Schönleber et al. (2014)
    R_neg: List[float] = []
    R_pos: List[float] = []
    name: str
    value: float
    for name, value in params.valuesdict().items():
        if not name.startswith("R"):
            continue
        elif name == "R0":
            continue
        if value < 0:
            R_neg.append(abs(value))
        else:
            R_pos.append(value)
    neg_sum: float = sum(R_neg)
    pos_sum: float = sum(R_pos)
    if pos_sum == 0:
        return 0.0
    mu: float = 1.0 - neg_sum / pos_sum
    if mu < 0.0:
        return 0.0
    elif mu > 1.0:
        return 1.0
    return mu


def _elements_to_parameters(
    elements: ndarray, taus: ndarray, add_capacitance: bool
) -> Parameters:
    assert type(elements) is ndarray, elements
    assert type(taus) is ndarray, taus
    assert type(add_capacitance) is bool, add_capacitance
    parameters: Parameters = Parameters()
    R0: float
    R0, elements = elements[0], elements[1:]
    parameters.add("R_0", R0, vary=False)
    L: float
    L, elements = elements[-1], elements[:-1]
    C: float
    if add_capacitance:
        C, elements = 1 / elements[-1], elements[:-1]
    for i, (R, t) in enumerate(zip(elements, taus), start=1):
        parameters.add(f"R_{i}", R, vary=False)
        parameters.add(f"t_{i}", t, vary=False)
    if add_capacitance:
        i += 1
        parameters.add(f"C_{i}", C)
    i += 1
    parameters.add(f"L_{i}", L, vary=False)
    return parameters


def _complex_residual(
    params: Parameters, circuit: Circuit, freq: ndarray, Z_exp: ndarray, weight: ndarray
) -> ndarray:
    assert type(params) is Parameters, params
    assert type(circuit) is Circuit, circuit
    assert type(freq) is ndarray, freq
    assert type(Z_exp) is ndarray, Z_exp
    assert type(weight) is ndarray, weight
    circuit.set_parameters(_from_lmfit(params))
    Z_fit: ndarray = circuit.impedances(freq)
    return array(
        [
            (weight * (Z_exp.real - Z_fit.real)) ** 2,
            (weight * (Z_exp.imag - Z_fit.imag)) ** 2,
        ]
    )


def _cnls_test(args: tuple) -> Tuple[int, float, Circuit, float]:
    freq: ndarray
    Z_exp: ndarray
    weight: ndarray
    num_RC: int
    add_capacitance: bool
    add_inductance: bool
    method: str
    max_nfev: int
    (
        freq,
        Z_exp,
        weight,
        num_RC,
        add_capacitance,
        add_inductance,
        method,
        max_nfev,
    ) = args
    assert type(freq) is ndarray, freq
    assert type(Z_exp) is ndarray, Z_exp
    assert type(weight) is ndarray, weight
    assert issubdtype(type(num_RC), integer), num_RC
    assert type(add_capacitance) is bool, add_capacitance
    assert type(add_inductance) is bool, add_inductance
    assert type(method) is str, method
    assert issubdtype(type(max_nfev), integer), max_nfev
    w: ndarray = 2 * pi * freq
    taus: ndarray = _generate_time_constants(w, num_RC)
    circuit: Circuit = _generate_circuit(taus, add_capacitance, add_inductance, True)
    fit: MinimizerResult
    try:
        fit = minimize(
            _complex_residual,
            _to_lmfit(circuit),
            method,
            args=(
                circuit,
                freq,
                Z_exp,
                weight,
            ),
            max_nfev=None if max_nfev < 1 else max_nfev,
        )
    except Exception:
        raise FittingError(format_exc())
    circuit.set_parameters(_from_lmfit(fit.params))
    mu: float = _calculate_mu(fit.params)
    Z_fit: ndarray = circuit.impedances(freq)
    Xps: float = _calculate_pseudo_chisqr(Z_exp, Z_fit, weight)
    return (
        num_RC,
        mu,
        circuit,
        Xps,
    )


# A shared integer value, which represents the smallest number of RC elements that has resulted
# in a mu-value that is less than the chosen mu-criterion. If this value is negative, then an
# appropriate number of RC elements has not yet been found. However, if the value is greater than
# the number of RC elements currently being evaluated by the process accessing the value, then
# the process should continue performing the current fitting. Otherwise, any ongoing fitting should
# be terminated as soon as possible and no further attempts should be made with a greater number of
# RC elements.
pool_optimal_num_RC = None  # multiprocessing.Value


# Initializer for the pool of processes that are used when performing the complex variant of the
# linear Kramers-Kronig test.
def _pool_init(args):
    global pool_optimal_num_RC
    pool_optimal_num_RC = args


def _cnls_mu_process(args: tuple) -> Tuple[int, float, Optional[Circuit], float]:
    global pool_optimal_num_RC
    freq: ndarray
    Z_exp: ndarray
    weight: ndarray
    mu_criterion: float
    num_RC: int
    add_capacitance: bool
    add_inductance: bool
    method: str
    max_nfev: int
    (
        freq,
        Z_exp,
        weight,
        mu_criterion,
        num_RC,
        add_capacitance,
        add_inductance,
        method,
        max_nfev,
    ) = args
    assert type(freq) is ndarray, freq
    assert type(Z_exp) is ndarray, Z_exp
    assert type(weight) is ndarray, weight
    assert issubdtype(type(mu_criterion), floating), mu_criterion
    assert issubdtype(type(num_RC), integer), num_RC
    assert type(add_capacitance) is bool, add_capacitance
    assert type(add_inductance) is bool, add_inductance
    assert type(method) is str, method
    assert issubdtype(type(max_nfev), integer), max_nfev

    def exit_early(params: Parameters, i: int, _, *args, **kwargs):
        if 0 <= pool_optimal_num_RC.value < num_RC:  # type: ignore
            return True
        return None

    w: ndarray = 2 * pi * freq
    taus: ndarray = _generate_time_constants(w, num_RC)
    circuit: Circuit = _generate_circuit(taus, add_capacitance, add_inductance, True)
    if 0 <= pool_optimal_num_RC.value < num_RC:  # type: ignore
        return (num_RC, -1.0, None, -1.0)
    fit: MinimizerResult
    try:
        fit = minimize(
            _complex_residual,
            _to_lmfit(circuit),
            method,
            args=(
                circuit,
                freq,
                Z_exp,
                weight,
            ),
            max_nfev=None if max_nfev < 1 else max_nfev,
            iter_cb=exit_early,
        )
    except Exception:
        raise FittingError(format_exc())
    if 0 <= pool_optimal_num_RC.value < num_RC:  # type: ignore
        return (num_RC, -1.0, None, -1.0)
    circuit.set_parameters(_from_lmfit(fit.params))
    mu: float = _calculate_mu(fit.params)
    Z_fit: ndarray = circuit.impedances(freq)
    Xps: float = _calculate_pseudo_chisqr(Z_exp, Z_fit, weight)
    if mu < mu_criterion:
        with pool_optimal_num_RC.get_lock():  # type: ignore
            pool_optimal_num_RC.value = num_RC  # type: ignore
    return (
        num_RC,
        mu,
        circuit,
        Xps,
    )


def _generate_variable_matrices(
    w: ndarray, num_RC: int, taus: ndarray, add_capacitance: bool, abs_Z_exp: ndarray
) -> Tuple[ndarray, ndarray]:
    assert type(w) is ndarray, w
    assert issubdtype(type(num_RC), integer), num_RC
    assert type(taus) is ndarray, taus
    assert type(add_capacitance) is bool, add_capacitance
    assert type(abs_Z_exp) is ndarray, abs_Z_exp
    # Generate matrices with the following columns (top to bottom is left to right)
    # - R0, series resistance
    # - Ri, resistance in parallel with taus[i - 1] where 0 < i <= num_RC
    # - (C, optional series capacitance)
    # - L, series inductance
    if add_capacitance is True:
        a_re = zeros((w.size, num_RC + 3))
        a_im = zeros((w.size, num_RC + 3))
        # Series capacitance
        a_im[:, -2] = -1 / (w * abs_Z_exp)  # No real part
    else:
        a_re = zeros((w.size, num_RC + 2))
        a_im = zeros((w.size, num_RC + 2))
    # Series resistance
    a_re[:, 0] = 1 / abs_Z_exp  # No imaginary part
    # Series inductance
    a_im[:, -1] = w / abs_Z_exp  # No real part
    # RC elements
    for i, tau in enumerate(taus):
        a_re[:, i + 1] = (1 / (1 + 1j * w * tau)).real / abs_Z_exp
        a_im[:, i + 1] = (1 / (1 + 1j * w * tau)).imag / abs_Z_exp
    return (
        a_re,
        a_im,
    )


def _generate_circuit(
    taus: ndarray, add_capacitance: bool, add_inductance: bool, cnls: bool
) -> Circuit:
    assert type(taus) is ndarray, taus
    assert type(add_capacitance) is bool, add_capacitance
    assert type(add_inductance) is bool, add_inductance
    assert type(cnls) is bool, cnls
    cdc: List[str] = ["R{R=1}"]
    t: float
    for t in taus:
        cdc.append(f"K{{R=1,t={t}F}}")
    if add_capacitance is True:
        cdc.append("C{C=1e-6}")
    if add_inductance is True:
        cdc.append("L{L=1e-3}")
    circuit: Circuit = parse_cdc("".join(cdc))
    if cnls is True:
        for element in circuit.get_elements():
            for parameter in element.get_parameters():
                if isinstance(parameter, Connection):
                    continue
                element.set_lower_limit(parameter, -inf)  # type: ignore
                element.set_upper_limit(parameter, inf)  # type: ignore
    return circuit


def _real_test(
    a_re: ndarray,
    a_im: ndarray,
    Z_exp: ndarray,
    abs_Z_exp: ndarray,
    w: ndarray,
    freq: ndarray,
    taus: ndarray,
    add_capacitance: bool,
    circuit: Circuit,
):
    assert type(a_re) is ndarray, a_re
    assert type(a_im) is ndarray, a_im
    assert type(Z_exp) is ndarray, Z_exp
    assert type(abs_Z_exp) is ndarray, abs_Z_exp
    assert type(w) is ndarray, w
    assert type(freq) is ndarray, freq
    assert type(taus) is ndarray, taus
    assert type(add_capacitance) is bool, add_capacitance
    assert type(circuit) is Circuit, circuit
    # Fit using the real part
    elements: ndarray = pinv(a_re).dot(Z_exp.real / abs_Z_exp)
    # Fit using the imaginary part to fix the series inductance (and capacitance)
    a_im = zeros((w.size, 2))
    a_im[:, -1] = w / abs_Z_exp
    if add_capacitance:
        a_im[:, -2] = -1 / (w * abs_Z_exp)
        elements[-2] = 1e-18  # Nullifies the series capacitance without dividing by 0
    circuit.set_parameters(
        _from_lmfit(_elements_to_parameters(elements, taus, add_capacitance))
    )
    Z_fit: ndarray = circuit.impedances(freq)
    coefs: ndarray = pinv(a_im).dot((Z_exp.imag - Z_fit.imag) / abs_Z_exp)
    # Extract the corrected series inductance (and capacitance)
    if add_capacitance:
        elements[-2:] = coefs
    else:
        elements[-1] = coefs[-1]
    circuit.set_parameters(
        _from_lmfit(_elements_to_parameters(elements, taus, add_capacitance))
    )


def _imaginary_test(
    a_im: ndarray,
    Z_exp: ndarray,
    abs_Z_exp: ndarray,
    freq: ndarray,
    taus: ndarray,
    add_capacitance: bool,
    weight: ndarray,
    circuit: Circuit,
):
    assert type(a_im) is ndarray, a_im
    assert type(Z_exp) is ndarray, Z_exp
    assert type(abs_Z_exp) is ndarray, abs_Z_exp
    assert type(freq) is ndarray, freq
    assert type(taus) is ndarray, taus
    assert type(add_capacitance) is bool, add_capacitance
    assert type(weight) is ndarray, weight
    assert type(circuit) is Circuit, circuit
    # Fit using the imaginary part
    elements: ndarray = pinv(a_im).dot(Z_exp.imag / abs_Z_exp)
    # Estimate the series resistance
    circuit.set_parameters(
        _from_lmfit(_elements_to_parameters(elements, taus, add_capacitance))
    )
    Z_fit: ndarray = circuit.impedances(freq)
    elements[0] = array_sum(weight * (Z_exp.real - Z_fit.real)) / array_sum(weight)
    circuit.set_parameters(
        _from_lmfit(_elements_to_parameters(elements, taus, add_capacitance))
    )


def _complex_test(
    a_re: ndarray,
    a_im: ndarray,
    Z_exp: ndarray,
    abs_Z_exp: ndarray,
    taus: ndarray,
    add_capacitance: bool,
    circuit: Circuit,
):
    assert type(a_re) is ndarray, a_re
    assert type(a_im) is ndarray, a_im
    assert type(Z_exp) is ndarray, Z_exp
    assert type(abs_Z_exp) is ndarray, abs_Z_exp
    assert type(taus) is ndarray, taus
    assert type(add_capacitance) is bool, add_capacitance
    assert type(circuit) is Circuit, circuit
    # Fit using the complex impedance
    x: ndarray = inv(a_re.T.dot(a_re) + a_im.T.dot(a_im))
    y: ndarray = a_re.T.dot(Z_exp.real / abs_Z_exp) + a_im.T.dot(Z_exp.imag / abs_Z_exp)
    elements: ndarray = x.dot(y)
    circuit.set_parameters(
        _from_lmfit(_elements_to_parameters(elements, taus, add_capacitance))
    )


def _test_wrapper(args: tuple) -> Tuple[int, float, Circuit, float]:
    test: str
    freq: ndarray
    Z_exp: ndarray
    weight: ndarray
    num_RC: int
    add_capacitance: bool
    test, freq, Z_exp, weight, num_RC, add_capacitance = args
    assert type(test) is str, test
    assert type(freq) is ndarray, freq
    assert type(Z_exp) is ndarray, Z_exp
    assert type(weight) is ndarray, weight
    assert issubdtype(type(num_RC), integer), num_RC
    assert type(add_capacitance) is bool, add_capacitance
    abs_Z_exp: ndarray = abs(Z_exp)
    w: ndarray = 2 * pi * freq
    taus: ndarray = _generate_time_constants(w, num_RC)
    a_re: ndarray
    a_im: ndarray
    a_re, a_im = _generate_variable_matrices(
        w, num_RC, taus, add_capacitance, abs_Z_exp
    )
    circuit: Circuit = _generate_circuit(taus, add_capacitance, True, False)
    # Solve the set of linear equations and update the circuit's parameters
    if test == "real":
        _real_test(
            a_re, a_im, Z_exp, abs_Z_exp, w, freq, taus, add_capacitance, circuit
        )
    elif test == "imaginary":
        _imaginary_test(
            a_im, Z_exp, abs_Z_exp, freq, taus, add_capacitance, weight, circuit
        )
    elif test == "complex":
        _complex_test(a_re, a_im, Z_exp, abs_Z_exp, taus, add_capacitance, circuit)
    # Calculate return values
    mu: float = _calculate_mu(_to_lmfit(circuit))
    Z_fit: ndarray = circuit.impedances(freq)
    Xps: float = _calculate_pseudo_chisqr(Z_exp, Z_fit, weight)
    return (
        num_RC,
        mu,
        circuit,
        Xps,
    )


def perform_test(
    data: DataSet,
    test: str = "complex",
    num_RC: int = 0,
    mu_criterion: float = 0.85,
    add_capacitance: bool = False,
    add_inductance: bool = False,
    method: str = "leastsq",
    max_nfev: int = -1,
    num_procs: int = -1,
) -> TestResult:
    """
    Performs a linear Kramers-Kronig test as described by Boukamp (1995).
    The results can be used to check the validity of an impedance spectrum before performing equivalent circuit fitting.
    If the number of RC elements is less than two, then a suitable number of RC elements is determined using the procedure described by Schönleber et al. (2014) based on a criterion for the calculated mu-value (zero to one).
    A mu-value of one represents underfitting and a mu-value of zero represents overfitting.

    References:

    - B.A. Boukamp, 1995, J. Electrochem. Soc., 142, 1885-1894 (https://doi.org/10.1149/1.2044210)
    - M. Schönleber, D. Klotz, and E. Ivers-Tiffée, 2014, Electrochim. Acta, 131, 20-27 (https://doi.org/10.1016/j.electacta.2014.01.034)

    Parameters
    ----------
    data: DataSet
        The data set to be tested.

    test: str = "complex"
        Supported values include "cnls", "complex", "imaginary", and "real". The "cnls" test performs a complex non-linear least squares fit using lmfit.minimize, which usually provides a good fit but is also quite slow.
        The "complex", "imaginary", and "real" tests perform the complex, imaginary, and real tests, respectively, according to Boukamp (1995).

    num_RC: int = 0
        The number of RC elements to use.
        A value greater than or equal to one results in the specific number of RC elements being tested.
        A value less than one results in the use of the procedure described by Schönleber et al. (2014) based on the chosen mu-criterion.
        If the provided value is negative, then the maximum number of RC elements to test is equal to the absolute value of the provided value.
        If the provided value is zero, then the maximum number of RC elements to test is equal to the number of frequencies in the data set.

    mu_criterion: float = 0.85
        The chosen mu-criterion. See Schönleber et al. (2014) for more information.

    add_capacitance: bool = False
        Add an additional capacitance in series with the rest of the circuit.

    add_inductance: bool = False
        Add an additional inductance in series with the rest of the circuit.
        Applies only to the "cnls" test.

    method: str = "leastsq"
        The fitting method to use when performing a "cnls" test.
        See the list of methods that are listed in the documentation for the lmfit package.
        Methods that do not require providing bounds for all parameters or a function to calculate the Jacobian should work.

    max_nfev: int = -1
        The maximum number of function evaluations when fitting.
        A value less than one equals no limit.
        Applies only to the "cnls" test.

    num_procs: int = -1
        The maximum number of parallel processes to use when performing a test.
        A value less than one results in using the number of cores returned by `multiprocessing.cpu_count`.
        Applies only to the "cnls" test.

    Returns
    -------
    TestResult
    """
    assert isinstance(data, DataSet), (
        type(data),
        data,
    )
    assert data.get_num_points() > 0
    assert type(test) is str, (
        type(test),
        test,
    )
    assert issubdtype(type(num_RC), integer), (
        type(num_RC),
        num_RC,
    )
    assert num_RC <= data.get_num_points(), f"{num_RC=} > {data.get_num_points()=}"
    assert issubdtype(type(mu_criterion), floating), (
        type(mu_criterion),
        mu_criterion,
    )
    if num_RC <= 0:
        assert mu_criterion >= 0.0 and mu_criterion <= 1.0, mu_criterion
    assert type(add_capacitance) is bool, (
        type(add_capacitance),
        add_capacitance,
    )
    assert type(add_inductance) is bool, (
        type(add_inductance),
        add_inductance,
    )
    assert type(method) is str, (
        type(method),
        method,
    )
    assert issubdtype(type(max_nfev), integer), (
        type(max_nfev),
        max_nfev,
    )
    assert issubdtype(type(num_procs), integer), (
        type(num_procs),
        num_procs,
    )
    if num_procs < 1:
        num_procs = cpu_count()
    freq: ndarray = data.get_frequency()
    Z_exp: ndarray = data.get_impedance()
    weight: ndarray = _generate_weight(Z_exp)
    params: Parameters
    mu: float
    circuit: Optional[Circuit] = None
    Xps: float  # pseudo chi-squared
    func: Callable
    fits: List[Tuple[int, float, Circuit, float]] = []
    progress_message: str = "Preparing arguments"
    progress.update_every_N_percent(0, message=progress_message)
    i: int
    num_RCs: List[int]
    if test == "cnls":
        if num_RC > 0:
            # Perform the test with a specific number of RC elements
            args = [
                (
                    freq,
                    Z_exp,
                    weight,
                    num_RC,
                    add_capacitance,
                    add_inductance,
                    method,
                    max_nfev,
                )
            ]
            progress_message = "Performing test"
            progress.update_every_N_percent(0, message=progress_message)
            try:
                if num_procs > 1:
                    # To prevent the GUI thread of DearEIS from locking up
                    with Pool(1) as pool:
                        fits = pool.map(_cnls_test, args)
                else:
                    fits = list(map(_cnls_test, args))
            except Exception:
                raise FittingError(format_exc())
            progress.update_every_N_percent(1, message=progress_message)
            num_RC, mu, circuit, Xps = fits[0]
        else:
            num_RC = abs(num_RC) or len(freq)
            num_RCs = list(range(1, num_RC + 1))
            # Find an appropriate number of RC elements based on the calculated mu-value and the
            # provided threshold value. Use multiple parallel processes if possible.
            args = (
                (
                    freq,
                    Z_exp,
                    weight,
                    mu_criterion,
                    num_RC,
                    add_capacitance,
                    add_inductance,
                    method,
                    max_nfev,
                )
                for num_RC in num_RCs
            )
            progress_message = "Performing test(s)"
            progress.update_every_N_percent(0, message=progress_message)
            try:
                if num_procs > 1:
                    with Pool(
                        num_procs,
                        initializer=_pool_init,
                        initargs=(Value("i", -1),),
                    ) as pool:
                        for i, res in enumerate(
                            pool.imap_unordered(_cnls_mu_process, args)
                        ):
                            progress.update_every_N_percent(
                                i + 1,
                                total=len(num_RCs),
                                message=progress_message,
                            )
                            if res[2] is not None:
                                fits.append(res)
                else:
                    _pool_init(Value("i", -1))
                    for i, res in enumerate(map(_cnls_mu_process, args)):
                        progress.update_every_N_percent(
                            i + 1,
                            total=len(num_RCs),
                            message=progress_message,
                        )
                        if res[2] is not None:
                            fits.append(res)
            except Exception:
                raise FittingError(format_exc())
            for (num_RC, mu, circuit, Xps) in sorted(fits, key=lambda _: _[0]):
                if mu < mu_criterion:
                    break
    else:
        supported_tests: List[str] = [
            "complex",
            "real",
            "imaginary",
        ]
        if test not in supported_tests:
            raise FittingError(f"Unsupported test: '{test}'")
        if num_RC > 0:
            num_RCs = [num_RC]
        else:
            num_RC = abs(num_RC) or len(freq)
            num_RCs = list(range(1, num_RC + 1))
        args = (
            (
                test,
                freq,
                Z_exp,
                weight,
                num_RC,
                add_capacitance,
            )
            for num_RC in num_RCs
        )
        progress_message = "Performing test(s)"
        progress.update_every_N_percent(0, message=progress_message)
        try:
            for i, res in enumerate(map(_test_wrapper, args)):
                progress.update_every_N_percent(
                    i + 1,
                    total=len(num_RCs),
                    message=progress_message,
                )
                fits.append(res)
        except Exception:
            raise FittingError(format_exc())
        if len(fits) == 1:
            num_RC, mu, circuit, Xps = fits[0]
        else:
            for (num_RC, mu, circuit, Xps) in sorted(fits, key=lambda _: _[0]):
                if mu < mu_criterion:
                    break
    # ========== Result ==========
    assert circuit is not None
    Z_fit: ndarray = circuit.impedances(freq)
    return TestResult(
        circuit,
        num_RC,
        mu,
        Xps,
        freq,
        Z_fit,
        # Residuals calculated according to eqs. 15 and 16
        # in Schönleber et al. (2014)
        *_calculate_residuals(Z_exp, Z_fit),
    )


def perform_exploratory_tests(
    data: DataSet,
    test: str = "complex",
    num_RCs: List[int] = [],
    mu_criterion: float = 0.85,
    add_capacitance: bool = False,
    add_inductance: bool = False,
    method: str = "leastsq",
    max_nfev: int = -1,
    num_procs: int = -1,
) -> List[TestResult]:
    """
    Performs a batch of linear Kramers-Kronig tests, which are then scored and sorted from best to worst before they are returned.

    Parameters
    ----------
    data: DataSet
        The data set to be tested.

    test: str = "complex"
        See `perform_test` for details.

    num_RCs: List[int] = []
        A list of integers representing the various number of RC elements to test.
        An empty list results in all possible numbers of RC elements up to the total number of frequencies being tested.

    mu_criterion: float = 0.85
        See `perform_test` for details.

    add_capacitance: bool = False
        See `perform_test` for details.

    add_inductance: bool = False
        See `perform_test` for details.

    method: str = "leastsq"
        See `perform_test` for details.

    max_nfev: int = -1
        See `perform_test` for details.

    num_procs: int = -1
        See `perform_test` for details.

    Returns
    -------
    List[TestResult]
    """
    assert isinstance(data, DataSet), (
        type(data),
        data,
    )
    assert data.get_num_points() > 0
    assert type(test) is str, (
        type(test),
        test,
    )
    assert type(num_RCs) is list, (
        type(num_RCs),
        num_RCs,
    )
    assert all(map(lambda _: issubdtype(type(_), integer), num_RCs))
    if len(num_RCs) > 0:
        assert (
            max(num_RCs) <= data.get_num_points()
        ), f"{max(num_RCs)=} > {data.get_num_points()=}"
    assert issubdtype(type(mu_criterion), floating), (
        type(mu_criterion),
        mu_criterion,
    )
    assert mu_criterion >= 0.0 and mu_criterion <= 1.0, mu_criterion
    assert type(add_capacitance) is bool, (
        type(add_capacitance),
        add_capacitance,
    )
    assert type(add_inductance) is bool, (
        type(add_inductance),
        add_inductance,
    )
    assert type(method) is str, (
        type(method),
        method,
    )
    assert issubdtype(type(max_nfev), integer), (
        type(max_nfev),
        max_nfev,
    )
    assert issubdtype(type(num_procs), integer), (
        type(num_procs),
        num_procs,
    )
    results: List[TestResult] = []
    if num_procs < 1:
        num_procs = cpu_count()
    freq: ndarray = data.get_frequency()
    Z_exp: ndarray = data.get_impedance()
    weight: ndarray = _generate_weight(Z_exp)
    if len(num_RCs) == 0:
        num_RCs = list(range(1, len(freq)))
    params: Parameters
    num_RC: int
    func: Callable
    progress_message: str = "Preparing arguments"
    progress.update_every_N_percent(0, message=progress_message)
    if test == "cnls":
        func = _cnls_test
        args = (
            (
                freq,
                Z_exp,
                weight,
                num_RC,
                add_capacitance,
                add_inductance,
                method,
                max_nfev,
            )
            for num_RC in num_RCs
        )
    else:
        supported_tests: List[str] = [
            "complex",
            "real",
            "imaginary",
        ]
        if test not in supported_tests:
            raise FittingError(f"Unsupported test: '{test}'")
        func = _test_wrapper
        args = (
            (
                test,
                freq,
                Z_exp,
                weight,
                num_RC,
                add_capacitance,
            )
            for num_RC in num_RCs
        )
    fits: List[Tuple[int, float, Circuit, float]] = []
    progress_message = "Performing test(s)"
    progress.update_every_N_percent(0, message=progress_message)
    try:
        if test == "cnls" and num_procs > 1:
            with Pool(num_procs) as pool:
                for i, res in enumerate(pool.imap_unordered(func, args)):
                    progress.update_every_N_percent(
                        i + 1,
                        total=len(num_RCs),
                        message=progress_message,
                    )
                    fits.append(res)
        else:
            for i, res in enumerate(map(func, args)):
                progress.update_every_N_percent(
                    i + 1,
                    total=len(num_RCs),
                    message=progress_message,
                )
                fits.append(res)
    except Exception:
        raise FittingError(format_exc())
    mu: float
    circuit: Circuit
    Xps: float
    for (num_RC, mu, circuit, Xps) in fits:
        Z_fit: ndarray = circuit.impedances(freq)
        results.append(
            TestResult(
                circuit,
                num_RC,
                mu,
                Xps,
                freq,
                Z_fit,
                # Residuals calculated according to eqs. 15 and 16
                # in Schönleber et al. (2014)
                *_calculate_residuals(Z_exp, Z_fit),
            )
        )
    results.sort(
        key=lambda _: _.calculate_score(mu_criterion),
        reverse=True,
    )
    return results



# pyimpspec is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 pyimpspec developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of pyimpspec's dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import (
    List,
    Union,
    Tuple,
)
from .base import (
    Connection,
    Element,
)


class Parallel(Connection):
    """
    Elements connected in parallel.

    Parameters
    ----------
    elements: List[Union[Element, Connection]]
        List of elements (and connections) that are connected in parallel.
    """

    def __init__(self, elements: List[Union[Element, Connection]]):
        super().__init__(elements)

    def to_stack(self, stack: List[Tuple[str, Union[Element, Connection]]]):
        stack.append(
            (
                "(",
                self,
            )
        )
        for element in reversed(self._elements):
            if isinstance(element, Connection):
                element.to_stack(stack)
            else:
                stack.append(
                    (
                        element.to_string(),
                        element,
                    )
                )
        stack.append(
            (
                ")",
                self,
            )
        )

    def to_string(self, decimals: int = -1):
        return (
            "("
            + "".join(
                map(lambda _: _.to_string(decimals=decimals), reversed(self._elements))
            )
            + ")"
        )

    def get_label(self) -> str:
        return "Parallel"

    def impedance(self, f: float) -> complex:
        if self._elements:
            return 1 / sum(map(lambda _: 1 / _.impedance(f), self._elements))  # type: ignore
        return complex(0, 0)

    def _str_expr(self, substitute: bool = False) -> str:
        if not self._elements:
            return "0"
        string: str = ""
        for element in reversed(self._elements):
            elem_str: str = element._str_expr(substitute=substitute)
            if string == "":
                string = f"1 / ({elem_str})"
            else:
                string += f" + 1 / ({elem_str})"
        return f"1 / ({string})"



# pyimpspec is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 pyimpspec developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of pyimpspec's dependencies and/or sources of portions of code are included in
# the LICENSES folder.

# This module uses Tikhonov regularization and non-negative least squares
# DRT-python-code commit: 9663ed8b331f521a9fcdb0b58fb2b34693df938c

from typing import (
    List,
)
from numpy import (
    array,
    float64,
    floating,
    identity,
    integer,
    int64,
    issubdtype,
    log as ln,
    logspace,
    ndarray,
    pi,
    sqrt,
    sum as array_sum,
    zeros,
)
from scipy.optimize import nnls
from pyimpspec.data import DataSet
from pyimpspec.analysis.fitting import _calculate_residuals
from pyimpspec.analysis.drt.result import (
    DRTResult,
    _calculate_chisqr,
)
import pyimpspec.progress as progress


_MODES: List[str] = ["real", "imaginary"]


def _generate_tikhonov_matrix(A: ndarray, I: ndarray, lambda_value: float) -> ndarray:
    return (A.T @ A) + lambda_value * I


def _solve(A: ndarray, b: ndarray) -> ndarray:
    return nnls(A, b)[0]


def _suggest_lambda(
    lambda_values: ndarray,
    solution_norms: ndarray,
) -> float:
    a: ndarray = zeros(lambda_values.size - 1, dtype=float64)
    for i in range(0, lambda_values.size - 1):
        a[i] = solution_norms[i] - solution_norms[i + 1]
    b: ndarray = zeros(a.size - 1, dtype=float64)
    for i in range(0, b.size - 1):
        b[i] = a[i] - a[i + 1]
    c: float
    for i, c in reversed(list(enumerate(b))):
        if c < 0.0:
            return lambda_values[(i + 1) if i < lambda_values.size - 2 else i]
    return lambda_values[-1]


def _calculate_drt_tr_nnls(
    data: DataSet,
    mode: str,
    lambda_value: float,
    num_procs: int,
) -> DRTResult:
    assert (
        hasattr(data, "get_frequency")
        and callable(data.get_frequency)
        and hasattr(data, "get_impedance")
        and callable(data.get_impedance)
    ), data
    assert (
        type(mode) is str and mode in _MODES
    ), f"Valid mode values: {', '.join(_MODES)}"
    assert issubdtype(type(lambda_value), floating), lambda_value
    assert issubdtype(type(num_procs), integer), num_procs
    # This is simply to avoid locking up the GUI thread in DearEIS.
    progress_message: str = "Preparing matrices"
    num_matrices: int = 4
    progress.update_every_N_percent(0, total=num_matrices, message=progress_message)
    i: int
    f: ndarray = data.get_frequency()
    Z: ndarray = data.get_impedance()
    omega: ndarray = 2 * pi * f
    tau: ndarray = 1 / f
    ln_tau: ndarray = ln(tau)
    delta_ln_tau: ndarray = zeros(omega.size, dtype=float64)
    for i in range(1, omega.size - 1):
        delta_ln_tau[i] = 0.5 * (ln_tau[i + 1] - ln_tau[i - 1])
    delta_ln_tau[0] = 0.5 * (ln_tau[1] - ln_tau[0])
    delta_ln_tau[-1] = 0.5 * (ln_tau[-1] - ln_tau[-2])
    progress.update_every_N_percent(1, total=num_matrices, message=progress_message)
    # Subtract the high-frequency resistance, which will be added back later,
    # and then normalize the impedances.
    R_inf: float = Z[0].real
    Z -= R_inf
    R_pol: float = Z[-1].real - Z[0].real
    Z_norm: ndarray = array(
        list(map(lambda _: complex(_.real / R_pol, _.imag / R_pol), Z))
    )
    # Added back so that the residuals and chi-square calculations will be correct.
    Z += R_inf
    # Prepare matrices and vectors
    I: ndarray = identity(omega.size, dtype=int64)
    progress.update_every_N_percent(2, total=num_matrices, message=progress_message)
    A: ndarray = zeros(
        (
            omega.size,
            omega.size,
        ),
        dtype=float64,
    )
    product: ndarray
    for i in range(0, omega.size):
        product = omega[i] * tau
        A[i, :] = (
            (product if mode == "imaginary" else 1) * delta_ln_tau / (1 + product**2)
        )
    progress.update_every_N_percent(3, total=num_matrices, message=progress_message)
    b: ndarray = A.T @ (Z_norm.real if mode == "real" else -Z_norm.imag)
    progress.update_every_N_percent(4, total=num_matrices, message=progress_message)
    progress_message = "Calculating DRT"
    progress.update_every_N_percent(0, message=progress_message)
    A_tikh: ndarray
    g_tau: ndarray
    if lambda_value <= 0.0:
        # Determine suitable regularization parameter if one hasn't been provided.
        min_log_lambda: int = -15
        max_log_lambda: int = 0
        lambda_values = logspace(
            min_log_lambda,
            max_log_lambda,
            # If the number of points is too high, then there is a risk of suggesting
            # a lambda value that is unnecessarily high.
            num=(max_log_lambda - min_log_lambda) * 1 + 1,
        )
        A_tikhs: List[ndarray] = []
        g_taus: List[ndarray] = []
        # residuals: ndarray = zeros(lambda_values.size, dtype=float64)
        solution_norms: ndarray = zeros(lambda_values.size, dtype=float64)
        for i, lambda_value in enumerate(lambda_values):
            A_tikhs.append(_generate_tikhonov_matrix(A, I, lambda_value))
            g_taus.append(_solve(A_tikhs[-1], b))
            # residuals[i] = sqrt(array_sum(((A.T @ A) @ g_taus[-1] - b) ** 2))
            solution_norms[i] = sqrt(array_sum(g_taus[-1] ** 2))
            progress.update_every_N_percent(
                i + 1,
                total=len(lambda_values),
                message=progress_message,
            )
        lambda_value = _suggest_lambda(lambda_values, solution_norms)
        A_tikh = A_tikhs[-1]
        g_tau = g_taus[-1]
        for i, lv in enumerate(lambda_values):
            if lv == lambda_value:
                A_tikh = A_tikhs[i]
                g_tau = g_taus[i]
                break
    else:
        A_tikh = _generate_tikhonov_matrix(A, I, lambda_value)
        g_tau = _solve(A_tikh, b)
        progress.update_every_N_percent(1, message=progress_message)
    R_pol_synthetic: float = array_sum(g_tau * delta_ln_tau)  # Should be (close to) 1.0
    # W: ndarray = A_tikh @ g_tau
    # res: float = sqrt(array_sum((W - b) ** 2))
    # lhs: float = sqrt(array_sum(W**2))
    Z_re_im: ndarray = zeros(omega.size, dtype=float64)
    for i in range(0, omega.size):
        product = omega[i] * tau
        Z_re_im[i] = array_sum(
            delta_ln_tau
            * ((product if mode == "imaginary" else 1) * g_tau / (1 + product**2))
        )
    Z_re_im = R_pol * Z_re_im
    Z_fit: ndarray
    if mode == "real":
        Z_fit = array(list(map(lambda _: complex(*_), zip(Z_re_im + R_inf, Z.imag))))
    else:
        Z_fit = array(list(map(lambda _: complex(*_), zip(Z.real, -Z_re_im))))
    gamma: ndarray = g_tau * R_pol
    return DRTResult(
        "TR-NNLS",
        tau,
        gamma,
        f,
        Z_fit,
        *_calculate_residuals(Z, Z_fit),
        # "tr-rbf" method with credible_intervals
        array([]),  # Mean
        array([]),  # Lower bound
        array([]),  # Upper bound
        # "bht" method
        array([]),  # Imaginary gamma
        {},
        # Stats
        _calculate_chisqr(Z, Z_fit),
        lambda_value,
    )



# pyimpspec is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 pyimpspec developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of pyimpspec's dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from collections import OrderedDict
from re import sub
from typing import (
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)
from numpy import (
    array,
    inf,
    integer,
    issubdtype,
    ndarray,
)
from sympy import (
    Expr,
    latex,
    sympify,
)


class Element:
    def __init__(self, keys: List[str]):
        self._label: str = ""
        self._fixed_parameters: Dict[str, bool] = {}
        self._lower_limits: Dict[str, float] = {}
        self._upper_limits: Dict[str, float] = {}
        # Make sure the keys exist in the relevant dictionaries
        key: str
        for key in keys:
            self._fixed_parameters[key] = False
            self._lower_limits[key] = -inf
            self._upper_limits[key] = inf
        self._identifier: int = -1

    def __repr__(self) -> str:
        return f"{self.get_label()} ({hex(id(self))})"

    @classmethod
    def get_extended_description(Class) -> str:
        assert (
            hasattr(Class, "__doc__")
            and type(Class.__doc__) is str
            and Class.__doc__.strip() != ""
        )
        return "\n".join(map(str.strip, Class.__doc__.split("\n"))).strip()

    @staticmethod
    def get_description() -> str:
        """
        Get a brief description of the element and its symbol.

        Returns
        -------
        str
        """
        raise Exception("Method has not been implemented!")

    @staticmethod
    def get_defaults() -> Dict[str, float]:
        """
        Get the default values for the element's parameters.

        Returns
        -------
        Dict[str, float]
        """
        raise Exception("Method has not been implemented!")

    @staticmethod
    def get_default_fixed() -> Dict[str, bool]:
        """
        Get whether or not the element's parameters are fixed by default.

        Returns
        -------
        Dict[str, bool]
        """
        raise Exception("Method has not been implemented!")

    @staticmethod
    def get_default_lower_limits() -> Dict[str, float]:
        """
        Get the default lower limits for the element's parameters.

        Returns
        -------
        Dict[str, float]
        """
        raise Exception("Method has not been implemented!")

    @staticmethod
    def get_default_upper_limits() -> Dict[str, float]:
        """
        Get the default upper limits for the element's parameters.

        Returns
        -------
        Dict[str, float]
        """
        raise Exception("Method has not been implemented!")

    def get_default_label(self) -> str:
        """
        Get the default label for this element.

        Returns
        -------
        str
        """
        if self._identifier >= 0:
            return f"{self.get_symbol()}_{self._identifier}"
        else:
            return self.get_symbol()

    def get_label(self) -> str:
        """
        Get the label assigned to a specific instance of the element.

        Returns
        -------
        str
        """
        if self._label == "":
            return self.get_default_label()
        return f"{self.get_symbol()}_{self._label}"

    def set_label(self, label: str) -> "Element":
        """
        Set the label assigned to a specific instance of the element.

        Parameters
        ----------
        label: str
            The new label.
        """
        assert type(label) is str, f"{label=}"
        self._label = label.strip()
        return self

    @staticmethod
    def get_symbol() -> str:
        """
        Get the symbol representing the element.

        Returns
        -------
        str
        """
        raise Exception("NOT YET IMPLEMENTED!")

    def to_string(self, decimals: int = -1) -> str:
        """
        Generates a string representation of the element.

        Parameters
        ----------
        decimals: int = -1
            The number of decimals used when formatting the current value and the limits for the element's parameters.
            -1 corresponds to no values being included in the output.

        Returns
        -------
        str
        """
        assert issubdtype(type(decimals), integer), decimals
        if decimals < 0:
            return self.get_symbol()
        parameters: List[str] = []
        for symbol, value in self.get_parameters().items():
            lower: float = self.get_lower_limit(symbol)
            upper: float = self.get_upper_limit(symbol)
            fixed: bool = self.is_fixed(symbol)
            string: str = f"{symbol}=" + (f"%.{decimals}E") % value
            if fixed:
                string += "F"
            elif lower != -inf and upper == inf:
                string += (f"/%.{decimals}E") % lower
            elif lower == -inf and upper != inf:
                string += (f"//%.{decimals}E") % upper
            elif lower != -inf and upper != inf:
                string += (f"/%.{decimals}E") % lower + (f"/%.{decimals}E") % upper
            parameters.append(string)
        cdc: str = self.get_symbol() + "{" + ",".join(parameters)
        if self._label != "":
            cdc += f":{self._label}"
        return cdc + "}"

    def _assign_identifier(self, current: int) -> int:
        """
        Get the internal identifier that is unique in the context of a circuit.
        Used internally when generating unique names for parameters when fitting a circuit to a
        data set.

        Parameters
        ----------
        current: int
            The most recently assigned identifier.
        """
        assert type(current) is int and current >= 0
        self._identifier = current
        return current + 1

    def get_identifier(self) -> int:
        """
        Get the internal identifier that is unique in the context of a circuit.
        Used internally when generating unique names for parameters when fitting a circuit to a
        data set.

        Returns
        -------
        int
        """
        return self._identifier

    def impedance(self, f: float) -> complex:
        """
        Calculates the complex impedance of the element at a specific frequency.

        Parameters
        ----------
        f: float
            Frequency in hertz.

        Returns
        -------
        complex
        """
        assert f > 0 and f < inf
        return complex(-999, 999)

    def impedances(self, freq: Union[list, ndarray]) -> ndarray:
        """
        Calculates the complex impedance of the element at specific frequencies.

        Parameters
        ----------
        freq: Union[list, ndarray]
            Frequencies in hertz.

        Returns
        -------
        ndarray
        """
        assert type(freq) is list or type(freq) is ndarray
        assert min(freq) > 0 and max(freq) < inf
        return array(list(map(self.impedance, freq)))

    def reset_parameters(self, keys: List[str]):
        """
        Resets the value, lower limit, upper limit, and fixed state of one or more parameters.

        Parameters
        ----------
        keys: List[str]
            Names of the parameters to reset.
        """
        assert type(keys) is list
        assert len(keys) > 0
        assert all(list(map(lambda _: type(_) is str, keys)))
        self.set_parameters({k: v for k, v in self.get_defaults().items() if k in keys})
        for k, v in self.get_default_lower_limits().items():
            if k not in keys:
                continue
            self.set_lower_limit(k, v)
        for k, v in self.get_default_upper_limits().items():
            if k not in keys:
                continue
            self.set_upper_limit(k, v)
        for k, v in self.get_default_fixed().items():
            if k not in keys:
                continue
            self.set_fixed(k, v)

    def get_parameters(self) -> "OrderedDict[str, float]":
        """
        Get the current parameters of the element.

        Returns
        -------
        OrderedDict[str, float]
        """
        # This implementation is just for basic testing
        return OrderedDict(
            {
                "A": -999,
                "B": 999,
            }
        )

    def set_parameters(self, parameters: Dict[str, float]):
        """
        Set new values for the parameters of the element.

        Parameters
        ----------
        parameters: Dict[str, float]
        """
        raise Exception("Method has not been implemented!")

    def is_fixed(self, key: str) -> bool:
        """
        Check if an element parameter should have a fixed value when fitting a circuit to a data
        set.
        True if fixed and False if not fixed.

        Parameters
        ----------
        key: str
            A key corresponding to an element parameter.

        Returns
        -------
        bool
        """
        assert type(key) is str and key.strip() != ""
        return self._fixed_parameters[key]

    def set_fixed(self, key: str, value: bool) -> "Element":
        """
        Set whether or not an element parameter should have a fixed value when fitting a circuit
        to a data set.

        Parameters
        ----------
        key: str
            A key corresponding to an element parameter.

        value: bool
            True if the value should be fixed.
        """
        assert type(key) is str and key.strip() != ""
        assert key in self._fixed_parameters, f"{key=}"
        assert type(value) is bool, f"{value=}"
        self._fixed_parameters[key] = value
        return self

    def get_lower_limit(self, key: str) -> float:
        """
        Get the lower limit for the value of an element parameter when fitting a circuit to a data
        set.
        The absence of a limit is represented by -numpy.inf.

        Parameters
        ----------
        key: str
            A key corresponding to an element parameter.

        Returns
        -------
        float
        """
        assert type(key) is str and key.strip() != ""
        return self._lower_limits[key]

    def set_lower_limit(self, key: str, value: float) -> "Element":
        """
        Set the upper limit for the value of an element parameter when fitting a circuit to a data
        set.

        Parameters
        ----------
        key: str
            A key corresponding to an element parameter.

        value: float
            The new limit for the element parameter. The limit can be removed by setting the limit
            to be -numpy.inf.
        """
        assert type(key) is str and key.strip() != ""
        assert key in self._lower_limits, f"{key=}"
        self._lower_limits[key] = float(value)
        return self

    def get_upper_limit(self, key: str) -> float:
        """
        Get the upper limit for the value of an element parameter when fitting a circuit to a data
        set.
        The absence of a limit is represented by numpy.inf.

        Parameters
        ----------
        key: str
            A key corresponding to an element parameter.

        Returns
        -------
        float
        """
        assert type(key) is str and key.strip() != ""
        return self._upper_limits[key]

    def set_upper_limit(self, key: str, value: float) -> "Element":
        """
        Set the upper limit for the value of an element parameter when fitting a circuit to a data
        set.

        Parameters
        ----------
        key: str
            A key corresponding to an element parameter.

        value: float
            The new limit for the element parameter. The limit can be removed by setting the limit
            to be numpy.inf.
        """
        assert type(key) is str and key.strip() != ""
        assert key in self._upper_limits, f"{key=}"
        self._upper_limits[key] = float(value)
        return self

    def _str_expr(self, substitute: bool = False) -> str:
        raise Exception("Method has not been implemented!")

    def _subs_str_expr(
        self, string: str, parameters: "OrderedDict[str, float]", symbols_only: bool
    ) -> str:
        assert type(string) is str
        assert type(parameters) is OrderedDict
        assert type(symbols_only) is bool
        k: str
        v: float
        for k, v in parameters.items():
            repl: str = f"{v:.12E}"
            if symbols_only:
                if self._label != "":
                    repl = f"{k}_{self._label}"
                elif self._identifier >= 0:
                    repl = f"{k}_{self._identifier}"
                else:
                    repl = f"{k}"
            pattern: str = r"(?<![a-zA-Z])" + k + r"(?![a-zA-Z])"
            string = sub(pattern, repl, string)
        return string

    def to_sympy(self, substitute: bool = False) -> Expr:
        assert type(substitute) is bool
        return sympify(self._str_expr(substitute=substitute))

    def to_latex(self) -> str:
        return f"Z = {latex(self.to_sympy(substitute=False))}"


class Connection:
    def __init__(self, elements: List[Union[Element, "Connection"]]):
        assert type(elements) is list
        assert all(
            map(lambda _: isinstance(_, Connection) or isinstance(_, Element), elements)
        )
        self._elements: List[Union[Element, "Connection"]] = elements

    def __repr__(self) -> str:
        return f"{self.get_label()} ({hex(id(self))})"

    def __len__(self) -> int:
        return len(self._elements)

    def __contains__(self, element: Union[Element, "Connection"]) -> bool:
        for ec in self.get_elements(flattened=False):
            if ec == element:
                return True
        return False

    def contains(
        self, element: Union[Element, "Connection"], top_level: bool = False
    ) -> bool:
        """
        Check if this connection contains a specific Element or Connection instance.

        Parameters
        ----------
        element: Union[Element, Connection]
            The Element or Connection instance to check for.

        top_level: bool = False
            Whether to only check in the current Connection instance instead of also checking in any nested Connection instances.

        Returns
        -------
        bool
        """
        if top_level:
            return element in self._elements
        return element in self

    def _assign_identifier(self, current: int) -> int:
        """
        Used internally when generating unique names for parameters when fitting a circuit to a
        data set.

        Parameters
        ----------
        current: int
            The most recently assigned identifier.

        Returns
        -------
        int
            The most recently assigned identifier.
        """
        element: Union[Element, "Connection"]
        for element in reversed(self._elements):
            current = element._assign_identifier(current)
        return current

    def get_label(self) -> str:
        raise Exception("Not yet implemented!")

    def to_string(self, decimals: int = -1) -> str:
        raise Exception("Not yet implemented!")

    def get_connections(self, flattened: bool = True) -> List["Connection"]:
        """
        Get the connections in this circuit.

        Parameters
        ----------
        flattened: bool = True
            Whether or not the connections should be returned as a list of all connections or as a list connections that may also contain more connections.

        Returns
        -------
        List[Connection]
        """
        if flattened:
            connections: List["Connection"] = []
            for element in self._elements:
                if isinstance(element, Connection):
                    connections.extend(
                        reversed(element.get_connections(flattened=flattened))
                    )
            return list(reversed(connections))
        return list(
            reversed(filter(lambda _: isinstance(_, Connection), self._elements))
        )

    def get_elements(
        self, flattened: bool = True
    ) -> List[Union[Element, "Connection"]]:
        """
        Get the elements in this circuit.

        Parameters
        ----------
        flattened: bool = True
            Whether or not the elements should be returned as a list of only elements or as a list of connections containing elements.

        Returns
        -------
        List[Union[Element, Connection]]
        """
        if flattened:
            elements: List[Union[Element, "Connection"]] = []
            for element in self._elements:
                if isinstance(element, Connection):
                    elements.extend(reversed(element.get_elements(flattened=flattened)))
                else:
                    elements.append(element)
            return list(reversed(elements))
        return list(reversed(self._elements))

    def substitute_element(self, ident: int, element: Element) -> bool:
        elem_con: Union[Element, Connection]
        for elem_con in self.get_elements(flattened=False):
            if isinstance(elem_con, Element):
                if elem_con.get_identifier() == ident:
                    element._assign_identifier(ident)
                    index: int = self._elements.index(elem_con)
                    self._elements.pop(index)
                    self._elements.insert(index, element)
                    return True
            else:
                if elem_con.substitute_element(ident, element) is True:
                    return True
        return False

    def impedance(self, f: float) -> complex:
        """
        Calculates the impedance of the connection at a single frequency.

        Parameters
        ----------
        f: float
            Frequency in Hz

        Returns
        -------
        complex
        """
        raise Exception("Method has not been implemented!")

    def impedances(self, freq: Union[list, ndarray]) -> ndarray:
        """
        Calculates the impedance of the element at multiple frequencies.

        Parameters
        ----------
        freq: Union[list, ndarray]
            Frequencies in Hz

        Returns
        -------
        ndarray
        """
        assert type(freq) is list or type(freq) is ndarray
        assert min(freq) > 0 and max(freq) < inf
        return array(list(map(self.impedance, freq)))

    def get_parameters(self) -> "Dict[int, OrderedDict[str, float]]":
        """
        Get the current element parameters of all elements nested inside this connection.
        The outer key is the unique identifier assigned to an element.
        The inner key is the symbol corresponding to an element parameter.

        Returns
        -------
        Dict[int, OrderedDict[str, float]]
        """
        parameters: "Dict[int, OrderedDict[str, float]]" = {}
        element: Union[Element, "Connection"]
        for element in self._elements:
            if isinstance(element, Connection):
                parameters.update(element.get_parameters())
                continue
            parameters[element.get_identifier()] = element.get_parameters()
        return parameters

    def set_parameters(self, parameters: Dict[int, Dict[str, float]]):
        """
        Set new element parameters to some/all elements nested inside this connection.

        Parameters
        ----------
        parameters: Dict[int, Dict[str, float]]
            The outer key is the unique identifier assigned to an element.
            The inner key is the symbol corresponding to an element parameter.
        """
        assert type(parameters) is dict
        element: Union[Element, "Connection"]
        for element in self._elements:
            if isinstance(element, Connection):
                element.set_parameters(parameters)
                continue
            ident: int = element.get_identifier()
            if ident not in parameters:
                continue
            element.set_parameters(parameters[ident])

    def get_element(self, ident: int) -> Optional[Element]:
        """
        Get a specific element based on its unique identifier.

        Parameters
        ----------
        ident: int
            The integer identifier that should be unique in the context of the circuit.

        Returns
        -------
        Optional[Element]
        """
        assert issubdtype(type(ident), integer), ident
        element: Optional[Union[Element, "Connection"]]
        for element in self._elements:
            if isinstance(element, Connection):
                element = element.get_element(ident)  # type: ignore
                if element is not None:
                    return element  # type: ignore
                continue
            elif element is not None and element.get_identifier() == ident:
                return element  # type: ignore
        return None

    def to_stack(self, stack: List[Tuple[str, Union[Element, "Connection"]]]):
        raise Exception("Method has not been implemented!")

    def _str_expr(self, substitute: bool = False) -> str:
        raise Exception("Method has not been implemented!")

    def to_sympy(self, substitute: bool = False) -> Expr:
        assert type(substitute) is bool
        return sympify(self._str_expr(substitute=substitute))

    def to_latex(self) -> str:
        return f"Z = {latex(self.to_sympy(substitute=False))}"
